import os
import schedule
from src.replifactory.loading import load_object
from src.replifactory import ExperimentWorker
_experiment = None


class Experiment:
    def __init__(self, directory=None, connect=False):
        global _experiment
        if "_experiment" in globals().keys():
            if _experiment is not None:
                print(_experiment)
                print("An experiment is already initialized in this thread: '%s'" % _experiment.directory)
                if _experiment.is_running():
                    print("Experiment '%s' is running!" % _experiment.directory)
                    _experiment.status()
                    # For non-interactive behaviour:
                    _experiment.stop()
                print("Running experiment STOPPED. New experiment initialized")

                # q = input("initialize new experiment?[y/n]")
                # if q == "y":
                #     _experiment = self
                # else:
                #     print("New experiment NOT initialized.")
                #     return
            else:
                _experiment = self
        else:
            _experiment = self
        self.directory = directory
        self.schedule = schedule.Scheduler()
        if not os.path.exists(directory):
            os.mkdir(directory)
            print("Created new experiment directory: '%s'" % directory)
        device_config_path = os.path.join(directory, "device_config.yaml")
        self._device = load_object(filepath=device_config_path)
        if self.device is not None:
            self.device.load()
            if connect:
                self.device.connect()
            else:
                print("Not connected to device.")
        if not os.path.exists(device_config_path):
            print("No device config found in '%s'" % self.directory)
        self.worker = ExperimentWorker(experiment=self)

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, dev):
        if dev is not None:
            if self.directory is not None:
                if os.path.exists(os.path.join(self.directory, "device_config.yaml")):
                    if dev.directory != self.directory:
                        raise RuntimeError("Device config %s already exists." % self.directory)
                        # q = input("Overwrite existing device config in '%s'? [y/n]" % self.directory)
                        # if q != "y":

        dev.directory = self.directory
        dev.connect()
        self._device = dev
        self.save()

    def save(self):
        self.device.save()

    def run(self):
        device = self.device
        assert not self.worker.is_alive()
        self.device.hard_stop_trigger = False
        self.device.soft_stop_trigger = False
        self.device.check_all()
        device.valves.close_all()
        device.stirrers.set_speed_all(speed=1)
        print("Closed valves, passed checks. starting background worker.")
        self.schedule.clear()
        self.schedule.every().minute.at(":57").do(device.thermometers.measure_temperature_background_thread)
        self.schedule.every().minute.at(":00").do(device.measure_od_all)
        self.schedule.every().minute.at(":30").do(device.update_cultures)
        self.schedule.every().day.at("11:00").do(device.flush_tubing)
        self.worker.start()

    def is_running(self):
        if not hasattr(self, "worker"):
            return False
        return self.worker.is_alive()

    def stop(self):
        self.worker.stop()

    def status(self, increase_verbosity=False):
        self.worker.status()
        if self.device is not None:
            for v in range(1, 8):
                if self.device.cultures[v] is not None:
                    self.device.cultures[v].show_parameters(increase_verbosity=increase_verbosity)
        else:
            print("Device not connected.")

    def plot(self, last_hours=24, increase_verbosity=False, td_ylim=None):
        for v in range(1, 8):
            if self.device.cultures[v] is not None:
                if self.device.cultures[v].is_active() or increase_verbosity:
                    fig = self.device.cultures[v].plot(last_hours=last_hours)
                    if td_ylim:
                        fig.axes[1].set_ylim(td_ylim[0], td_ylim[1])



import time
import os


def make_dilution(device, vial, pump1_volume=0, pump2_volume=0, pump3_volume=0, extra_vacuum=5):
    assert 0 <= pump1_volume <= 15
    assert 0 <= pump2_volume <= 15
    assert 0 <= pump3_volume <= 15
    assert 0 <= extra_vacuum <= 15
    device.locks_vials[vial].acquire()
    device.lock_pumps.acquire()
    assert not device.is_pumping()  # must have been checked before releasing the pump lock
    assert not device.hard_stop_trigger
    device.stirrers.set_speed(vial=vial, speed=2)
    time.sleep(1)
    device.valves.open(vial)  # valve number might be different for e.g. a lagoon setup

    if pump1_volume > 0:
        device.pump1.pump(pump1_volume)
        device.pump1.stock_volume -= pump1_volume
        device.cultures[vial]._time_last_dilution[1] = int(time.time())
    if pump2_volume > 0:
        device.pump2.pump(pump2_volume)
        device.pump2.stock_volume -= pump2_volume
        device.cultures[vial]._time_last_dilution[2] = int(time.time())

    if pump3_volume > 0:
        device.pump3.pump(pump3_volume)
        device.pump3.stock_volume -= pump3_volume
        device.cultures[vial]._time_last_dilution[3] = int(time.time())
    log_dilution(device=device, vial_number=vial, pump1_volume=pump1_volume,
                 pump2_volume=pump2_volume, pump3_volume=pump3_volume)
    while device.is_pumping():
        time.sleep(0.5)
        assert not device.hard_stop_trigger
    device.stirrers.set_speed(vial=vial, speed=1)
    dilution_volume = pump1_volume + pump2_volume + pump3_volume
    waste_volume = dilution_volume + extra_vacuum
    waste_volume = round(waste_volume, 2)

    device.pump4.pump(waste_volume)
    device.pump4.stock_volume -= dilution_volume
    device.cultures[vial]._time_last_dilution[4] = int(time.time())

    while device.pump4.is_busy():
        time.sleep(0.5)
        assert not device.hard_stop_trigger
    device.stirrers.set_speed(vial=vial, speed=2)
    assert not device.is_pumping()  # make sure before closing valves
    time.sleep(2)
    device.valves.close(valve=vial)
    device.locks_vials[vial].release()
    device.lock_pumps.release()
    log_dilution(device=device, vial_number=vial, pump4_volume=waste_volume)
    return 0


def log_dilution(device, vial_number, pump1_volume=0, pump2_volume=0, pump3_volume=0, pump4_volume=0):
    directory = os.path.join(device.directory, "vial_%d" % vial_number)
    if not os.path.exists(directory):
        os.mkdir(directory)
    filepath = os.path.join(directory, "dilutions.csv")
    if not os.path.exists(filepath):
        with open(filepath, "w+") as f:
            f.write("time,pump1_volume,pump2_volume,pump3_volume,pump4_volume\n")
    with open(filepath, "a") as f:
        data_string = str(int(time.time()))
        for vol in [pump1_volume, pump2_volume, pump3_volume, pump4_volume]:
            if vol == 0:
                vol_str = ""
            else:
                vol_str = str(vol)
            data_string += "," + vol_str
        data_string += "\n"
        f.write(data_string)



import time
import os
import pyftdi.i2c


class Thermometers:
    def __init__(self, device):
        self.device = device
        self.thermometer_vials = None
        self.thermometer_board = None

        if self.device.is_connected():
            self.connect()

    def connect(self):
        try:
            self.thermometer_vials = self.device.i2c.get_port(self.device.PORT_THERMOMETER_VIALS)
            self.thermometer_board = self.device.i2c.get_port(self.device.PORT_THERMOMETER_BOARD)
        except pyftdi.i2c.I2cNackError:
            print("PCA9685 PWM controller connection ERROR.")

    # near DC-DC converters which get warm

    def measure_temperature_background_thread(self):
        if self.device.od_worker.queue.empty():
            self.device.od_worker.queue.put(self.measure_temperature)
        else:
            print("Temperature measurement not queued. OD thread queue is not empty.")

    def measure_temperature(self):
        temps = []
        for thermometer_port in [self.thermometer_vials, self.thermometer_board]:
            thermometer_port.write([0x04])
            time.sleep(0.06)
            data = thermometer_port.read_from(0x04, 2)
            digital_temp = ((data[0] << 8) | data[1]) >> 4
            celsius_temp = digital_temp*0.0625
            temps += [celsius_temp]
        t_vials, t_board = temps
        self.log_temperature(t_vials=t_vials, t_board=t_board)
        return t_vials, t_board

    def log_temperature(self, t_vials, t_board):
        """
        logs temperature if device.directory exists
        """
        if os.path.exists(str(self.device.directory)):
            filepath = os.path.join(self.device.directory, "temperature.csv")
            if not os.path.exists(filepath):
                with open(filepath, "w+") as f:
                    f.write("time,temperature_vials,temperature_board\n")
            with open(filepath, "a") as f:
                f.write("%d,%.3f,%.3f\n" % (time.time(), t_vials, t_board))




import numpy as np


def split_bytes(value, n_bits, n_bytes):
    assert 0 <= value <= 1
    value = int(value*(2**n_bits-1))
    value = max(1, value)
    assert value < 2**n_bits

    bytes_list = [0b00]*n_bytes
    for i in range(n_bytes):
        lsb = value & ((1 << 8) - 1)
        value = value >> 8
        bytes_list[i] = lsb
    bytes_list = bytes_list[::-1]
    return bytes_list


class Stepper:
    REGISTER_ABS_POS = 0x01        # Current position, 22 bits
    REGISTER_EL_POS = 0x02         # Electrical position, 9 bits
    REGISTER_MARK = 0x03           # Mark position, 22 bits
    REGISTER_SPEED = 0x04          # Current speed, 20 bits
    REGISTER_ACC = 0x05            # Acceleration, 12 bits
    REGISTER_DEC = 0x06            # Deceleration, 12 bits
    REGISTER_MAX_SPEED = 0x07      # Maximum speed, 10 bits
    REGISTER_MIN_SPEED = 0x08      # Minimum speed, 13 bits
    REGISTER_FS_SPD = 0x15         # Full-step speed, 10 bits
    REGISTER_KVAL_HOLD = 0x09      # Holding KVAL, 8 bits
    REGISTER_KVAL_RUN = 0x0A       # Constant speed KVAL, 8 bits
    REGISTER_KVAL_ACC = 0x0B       # Acceleration starting KVAL, 8 bits
    REGISTER_KVAL_DEC = 0x0C       # Deceleration starting KVAL, 8 bits
    REGISTER_INT_SPEED = 0x0D      # Intersect speed, 14 bits
    REGISTER_ST_SLP = 0x0E         # Start slope, 8 bits
    REGISTER_FN_SLP_ACC = 0x0F     # Acceleration final slope, 8 bits
    REGISTER_FN_SLP_DEC = 0x10     # Deceleration final slope, 8 bits
    REGISTER_K_THERM = 0x11        # Thermal compensation factor, 4 bits
    REGISTER_ADC_OUT = 0x12        # ADC output, 5 bits
    REGISTER_OCD_TH = 0x13         # OCD threshold, 4 bits
    REGISTER_STALL_TH = 0x14       # STALL threshold, 7 bits
    REGISTER_STEP_MODE = 0x16      # Step mode, 8 bits
    REGISTER_ALARM_EN = 0x17       # Alarm enable, 8 bits
    REGISTER_CONFIG = 0x18         # IC configuration, 16 bits
    REGISTER_STATUS = 0x19         # Status, 16 bits

    min_speed_rps = 0.01
    max_speed_rps = 4
    acceleration = 0.01
    deceleration = 0.01
    full_step_speed = 0.1
    stall_threshold = 0.5
    _kval_hold = 0
    _kval_run = 0.8
    _kval_acc = 0.57
    _kval_dec = 0.3

    def __init__(self, device, cs):
        self.device = device
        self.cs = cs
        self.port = None
        self.step_mode = None
        if self.device.is_connected():
            self.connect()

    @property
    def kval_hold(self):
        return self._kval_hold

    @kval_hold.setter
    def kval_hold(self, value):
        assert 0 <= value <= 1
        self.write_register(self.REGISTER_KVAL_HOLD, value=value, n_bits=7, n_bytes=1)
        self._kval_hold=value

    @property
    def kval_acc(self):
        return self._kval_acc

    @kval_acc.setter
    def kval_acc(self, value):
        assert 0 <= value <= 1
        self.write_register(self.REGISTER_KVAL_ACC, value=value, n_bits=7, n_bytes=1)
        self._kval_acc = value


    @property
    def kval_dec(self):
        return self._kval_dec

    @kval_dec.setter
    def kval_dec(self, value):
        assert 0 <= value <= 1
        self.write_register(self.REGISTER_KVAL_DEC, value=value, n_bits=7, n_bytes=1)
        self._kval_dec = value

    @property
    def kval_run(self):
        return self._kval_run

    @kval_run.setter
    def kval_run(self, value):
        assert 0 <= value <= 1
        self.write_register(self.REGISTER_KVAL_RUN, value=value, n_bits=7, n_bytes=1)
        self._kval_run = value

    def connect(self):
        self.port = self.device.spi.get_port(cs=self.cs, freq=1e5, mode=3)
        # self.port.set_mode(3)
        self.write_register(self.REGISTER_FS_SPD, value=self.full_step_speed, n_bits=10, n_bytes=2)  # k value const
        self.write_register(self.REGISTER_STALL_TH, value=self.stall_threshold, n_bits=6, n_bytes=1)
        self.kval_hold=self._kval_hold
        self.kval_acc = self._kval_acc
        self.kval_run = self._kval_run
        # self.write_register(self.REGISTER_KVAL_HOLD, value=self.kval_hold, n_bits=7, n_bytes=1)
        # self.write_register(self.REGISTER_KVAL_RUN, value=self.kval_run, n_bits=7, n_bytes=1)
        # self.write_register(self.REGISTER_KVAL_ACC, value=self.kval_acc, n_bits=7, n_bytes=1)
        self.set_min_speed(rot_per_sec=self.min_speed_rps)
        self.set_max_speed(rot_per_sec=self.max_speed_rps)
        self.set_acceleration(value=self.acceleration)
        self.set_deceleration(value=self.deceleration)

    def set_acceleration(self, value=1e-4):
        self.write_register(self.REGISTER_ACC, value=value, n_bits=12, n_bytes=2)

    def set_deceleration(self, value=1e-3):
        self.write_register(self.REGISTER_DEC, value=value, n_bits=12, n_bytes=2)

    def set_min_speed(self, rot_per_sec=0.1):
        # steps_per_sec = integer * 2**-24/250e-9   # page 43 in L6470H datasheet
        correction_factor = 1.322
        rot_per_sec = rot_per_sec * correction_factor
        steps_per_sec = rot_per_sec * 2 ** 22 / 25600
        integer = steps_per_sec * 250e-9/2**-24
        self.write_register(self.REGISTER_MIN_SPEED, value=integer/2**12, n_bits=12, n_bytes=2)

    def set_max_speed(self, rot_per_sec=5.0):
        # steps_per_sec = integer * 2**-18/250e-9   # page 43 in L6470H datasheet
        correction_factor = 1.322
        rot_per_sec = rot_per_sec * correction_factor
        steps_per_sec = rot_per_sec * 2 ** 22 / 25600
        integer = steps_per_sec * 250e-9 / 2 ** -18
        self.write_register(self.REGISTER_MAX_SPEED, value=integer/2**10, n_bits=10, n_bytes=2)

    def write_register(self, reg, value, n_bits, n_bytes):
        set_param = reg | 0b00000000
        self.port.write([set_param])
        bytes_to_write = split_bytes(value, n_bits, n_bytes)
        for b in bytes_to_write:
            self.port.write([b])
        bytes_read = self.read_register(reg=reg, n_bytes=n_bytes)
        assert [bytes_to_write[i] == bytes_read[i] for i in range(n_bytes)]

    def read_register(self, reg, n_bytes=3):
        get_param = reg | 0b00100000
        self.port.write([get_param])
        res = []
        for b in range(n_bytes):
            res += self.port.read(1)
        return res

    def move(self, n_rotations=1):
        """
        move the given number of rotations
        By default uses 1/128 microstepping to minimize noise and vibration.
        If number of required microsteps can not fit in the register (> 2^22),
        the step mode is decreased to 1/64 or lower.
        """
        # if n_rotations <= 1:
        #     self.set_max_speed(rot_per_sec=0.1)
        # else:
        #     self.set_max_speed(rot_per_sec=self.max_speed_rps)

        if self.step_mode != 7:
            self.set_step_mode(7)
        assert not self.is_pumping()
        assert self.device.valves.not_all_closed()
        if n_rotations >= 0:
            direction_bit = 0b1
        else:
            direction_bit = 0b0

        move_header_byte = 0b01000000 | direction_bit
        max_n_microsteps = 2**22  # 22 bit

        steps_per_rotation = 200
        microsteps_per_step = 2 ** self.step_mode
        n_microsteps = abs(n_rotations) * steps_per_rotation * microsteps_per_step

        if n_microsteps > max_n_microsteps:
            required_step_mode = self.step_mode - 1 - int(np.log2(n_microsteps/max_n_microsteps))
            self.set_step_mode(required_step_mode)
            microsteps_per_step = 2 ** self.step_mode
            n_microsteps = abs(n_rotations) * steps_per_rotation * microsteps_per_step

        write_bytes = split_bytes(n_microsteps/max_n_microsteps, 22, 3)
        self.port.write([move_header_byte])
        for b in write_bytes:
            self.port.write([b])

    def run(self, speed=0.001):
        """
        run indefinitely at constant speed
        """
        if self.step_mode != 7:
            self.set_step_mode(7)

        if speed >= 0:
            direction_bit = 0b1
        else:
            direction_bit = 0b0
        speed = abs(speed)

        run_header_byte = 0b01010000 | direction_bit
        write_bytes = split_bytes(speed, 20, 3)
        self.port.write([run_header_byte])
        for b in write_bytes:
            self.port.write([b])

    def print_status(self):
        msb, lsb = self.read_register(self.REGISTER_STATUS, n_bytes=2)
        step_loss_a = not (msb >> 5 & 0b1)
        step_loss_b = not (msb >> 6 & 0b1)

        step_loss_string = ""
        if step_loss_a:
            step_loss_string += "stall on bridge a; "
        if step_loss_b:
            step_loss_string += "stall on bridge b; "
        if not (step_loss_b or step_loss_a):
            step_loss_string += "no step loss detected"

        overcurrent = not (msb >> 4 & 0b1)
        thermal_shutdown = not (msb >> 3 & 0b1)     # 160 deg C
        thermal_warning = not (msb >> 2 & 0b1)      # 130 deg C
        undervoltage_lockout = not (msb >> 1 & 0b1)
        status = lsb >> 5 & 0b11  # 00: stopped, 01:accelerating, 10:decelerating, 11:constant speed
        status_string = ["stopped", "accelerating", "decelerating", "moving at constant speed"][status]
        busy = not (lsb >> 1 & 0b1)
        high_impedance = bool(lsb & 0b1)
        print("Overcurrent:\t", overcurrent)
        print("Thermal shutdown:\t", thermal_shutdown)
        print("Thermal warning:\t", thermal_warning)
        print("Undervoltage lockout flag:\t", undervoltage_lockout)
        print("Status:\t", status_string)
        print("Busy:\t", busy)
        print("High impedance:\t", high_impedance)
        print("Step loss:\t", step_loss_string)
        self.port.write([0b11010000])  # GetStatus command, resets warning flags

    def is_busy(self):
        self.port.write([0b11010000])  # GetStatus command, resets warning flags
        msb, lsb = self.read_register(self.REGISTER_STATUS, n_bytes=2)
        busy = not (lsb >> 1 & 0b1)
        return busy

    def driver_is_responsive(self):
        self.port.write([0x19])  # GetStatus command, resets warning flags
        msb, lsb = self.read_register(self.REGISTER_STATUS, n_bytes=2)
        return not (msb == 255 and lsb == 255)

    def is_pumping(self):
        if not self.driver_is_responsive():
            return False
        self.port.write([0b11010000])  # GetStatus command, resets warning flags
        msb, lsb = self.read_register(self.REGISTER_STATUS, n_bytes=2)
        status = lsb >> 5 & 0b11
        return status > 0

    def is_hiz(self):
        msb, lsb = self.read_register(self.REGISTER_STATUS, n_bytes=2)
        hiz = bool(lsb & 0b1)
        return hiz

    def stop(self):
        """
        decelerate with programmed deceleration value until the MIN_SPEED value
        is reached and then stop the motor
        """
        self.port.write([0b10110000])

    def stop_hard(self):
        """
        stop the motor instantly, ignoring deceleration constraints
        """
        self.port.write([0b10111000])

    def hard_hiz(self):
        self.port.write([0b10101000])  # Hard HiZ

    def soft_hiz(self):
        self.port.write([0b10100000])  # soft HiZ

    def reset(self):
        self.port.write([0b11000000])
        self.__init__(device=self.device, cs=self.cs)

    def set_step_mode(self, mode=7):
        """
        See page 48 in L6470H datasheet

        0: Full-step
        1: Half-step
        2: 1/4 microstep
        3: 1/8 microstep
        4: 1/16 microstep
        5: 1/32 microstep
        6: 1/64 microstep
        7: 1/128 microstep
        """
        assert mode in [0, 1, 2, 3, 4, 5, 6, 7]
        assert not self.is_pumping()
        self.hard_hiz()
        self.port.write([self.REGISTER_STEP_MODE])
        self.port.write([mode])
        read_mode = self.read_register(self.REGISTER_STEP_MODE)[0]
        assert mode == read_mode
        self.step_mode = mode



import ipywidgets as widgets
from IPython.core.display import clear_output
from ipywidgets import VBox


class ExperimentStatusTab:
    title = "Experiment status"

    def __init__(self, main_gui):
        self.main_gui = main_gui
        self.button = widgets.Button(description="show status")
        self.output = widgets.Output()
        self.button.on_click(self.handle_button_action)
        self.widget = VBox([self.button, self.output])

    def handle_button_action(self, button):
        with self.output:
            clear_output()
            self.main_gui.experiment.status(increase_verbosity=True)



import os
import glob
import ipywidgets as widgets

from ipywidgets import Layout, VBox, HBox

from src.replifactory import get_ftdi_addresses
from IPython.display import clear_output

# class ExperimentWidget:
#     def __init__(self, status_bar):
#         self.status_bar = status_bar
#         print("output made")
#         with self.status_bar.output:
#             self.status_bar.main_gui.experiment = Experiment("NewExperiment")
#         # if self.status_bar.main_gui.experiment is not None:
#         #     with self.status_bar.output:
#         #         self.status_bar.main_gui.experiment.status()
# box_layout = Layout(display='flex',
#                     flex_flow='column',
#                     align_items='stretch',
#                     border='solid 1px gray', )


class StatusBar:
    config_paths = glob.glob("../**/device_config.yaml")
    config_paths = [os.path.relpath(p) for p in config_paths]
    experiment_directories = [os.path.relpath(os.path.join(p, "..")) for p in config_paths]
    layout_width = "350px"

    def __init__(self, main_gui):
        self.main_gui = main_gui
        ftdi_addresses = get_ftdi_addresses()
        ndev = len(ftdi_addresses)
        self.title = widgets.HTML('<b><font size=2>Status:</b><br>Devices available: %d' % ndev)
        self.subtitle = widgets.Output()
        self.output_title = widgets.HTML("Output:")
        self.output = widgets.Output()
        self.output.layout.width = self.layout_width
        self.widget = None
        self.check_button = widgets.Button(description="CHECK")
        self.run_button = widgets.Button(description="RUN", disabled=True)
        self.stop_button = widgets.Button(description="STOP", disabled=True)
        self.check_button.on_click(self.handle_check_button)
        self.run_button.on_click(self.handle_run_button)
        self.stop_button.on_click(self.handle_stop_button)
        self.update()

    # def ask_question(self, question):
    #     input_label = widgets.Label("How can i help you? ")
    #     input_text = widgets.Text()

    def handle_check_button(self, b):
        with self.output:
            clear_output()
            assert not self.main_gui.experiment.is_running()
            self.main_gui.experiment.device.check_all()
            self.run_button.disabled = False
            # self.update()

    def handle_run_button(self, b):
        with self.output:
            clear_output()
            self.main_gui.experiment.run()
            self.run_button.disabled = True
            self.stop_button.disabled = False
            # self.update()

    def handle_stop_button(self, b):
        with self.output:
            clear_output()
            self.main_gui.experiment.stop()
            self.stop_button.disabled = True
            self.run_button.disabled = False
            # self.update()

    def update(self):
        with self.subtitle:
            clear_output()
            try:
                print("    EXPERIMENT: %s" % self.main_gui.experiment.directory)
            except:
                print("    EXPERIMENT: None")
            try:
                if self.main_gui.device.__class__ is type(None):
                    print("  DEVICE CLASS: None")
                else:
                    print("  DEVICE CLASS: %s" % str(self.main_gui.device.__class__).split(".")[-1][:-2])
            except:
                print("  DEVICE CLASS: None")
            try:
                print("DEVICE ADDRESS: %s" % self.main_gui.device.ftdi_address)
            except:
                print("DEVICE ADDRESS: None")

        widget_list = [self.title, self.subtitle, self.check_button, HBox([self.run_button, self.stop_button]),
                       self.output_title, self.output]
        self.widget = VBox(widget_list, layout=Layout(width=self.layout_width))




import socket
from pushover import init, Client


class Notifier:
    def __init__(self, title=None, token=None, userkey=None):
        self.title = title
        init(token)
        ip = get_ip_address()
        if self.title is None:
            self.title = ip
        self.client = Client(userkey)
        self.client.send_message("My IP address:\n"+ip, title=self.title)

    def info_message(self, message):
        self.client.send_message(message, title=self.title + " info")

    def alert_message(self, message):
        self.client.send_message(message, title=self.title + " ALERT")


def get_ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
    s.close()
    return ip



import ipywidgets as widgets
from IPython.display import clear_output
from ipywidgets import HBox, VBox


class DeviceConfigTab:
    def __init__(self, main_gui):
        self.title = "Device config"
        self.main_gui = main_gui
        self.button = widgets.Button(description="Show device parameters")
        self.button.on_click(self.handle_button_click)
        self.output = widgets.Output()
        self.widget = VBox([self.button, self.output])
        self.update()

    def handle_button_click(self, button):
        self.update()

    def update(self):
        if self.main_gui.device is not None:
            with self.output:
                clear_output()
                self.main_gui.device.show_parameters()



import time
import pyftdi.i2c

class Photodiodes:
    def __init__(self, device):
        self.device = device
        self.adc_port = None
        self.multiplexer_port = None

        if self.device.is_connected():
            self.connect()

    def connect(self):
        try:
            self.adc_port = self.device.i2c.get_port(self.device.PORT_ADC)  # photodiodes ADC
            self.multiplexer_port = self.device.i2c.get_port(self.device.PORT_GPIO_MULTIPLEXER_ADC)
            self.multiplexer_port.write_to(6, [0x00])  # set all GPIO pins as output
            self.multiplexer_port.write_to(7, [0x00])  # set all GPIO pins as output
        except pyftdi.i2c.I2cNackError:
            print("Photodiode multiplexer or ADC connection ERROR.")

    def switch_to_vial(self, vial):
        vial = vial-1
        assert 0 <= vial <= 6
        self.multiplexer_port.write_to(3, [6 - vial])

    def measure(self, gain=8, bitrate=16, continuous_conversion=False):
        ready_bit = 0b10000000
        if continuous_conversion:
            conversion_mode = 0b00010000
        else:
            conversion_mode = 0b00000000  # One-shot conversion mode
        bitrate_bits = {12: 0b0000,
                        14: 0b0100,
                        16: 0b1000,
                        18: 0b1100}
        bitrate_to_samples_per_second = {12: 240,   # from MCP3421 datasheet
                                         14: 60,
                                         16: 15,
                                         18: 3.75}
        gain_bits = {1: 0b00,
                     2: 0b01,
                     4: 0b10,
                     8: 0b11}

        write_byte = ready_bit | conversion_mode | bitrate_bits[bitrate] | gain_bits[gain]
        self.adc_port.write([write_byte])

        new_data_is_ready = False
        for _ in range(100):  # try reading until conversion ready
            samples_per_second = bitrate_to_samples_per_second[bitrate]
            seconds_per_sample = 1/samples_per_second

            time.sleep(seconds_per_sample)
            response = self.adc_port.read(4)
            data = response[:2]
            config = response[3:]
            if bitrate == 18:
                data = response[:3]  # 3 bytes for 18bit

            new_data_is_ready = config[0] < 128  # if bit 7 (ready bit) is 1: NOT READY YET!
            if new_data_is_ready:
                break

        assert new_data_is_ready
        data_bits = "".join([bin(x)[2:].rjust(8, "0") for x in data])[-bitrate:]
        sign_bit = data_bits[0]
        digital_signal = sum([2 ** i for i in range(len(data_bits)) if data_bits[::-1][i] == "1"])
        if sign_bit == "1":
            digital_signal = digital_signal-2**bitrate
        millivolts = 2 * 2.048 / 2 ** bitrate * digital_signal * 1000 / gain
        millivolts = round(millivolts, 10)
        lsb_mv = 2.048 * 2 / 2 ** bitrate * 1000 / gain  # least significant bit millivolts
        return millivolts, lsb_mv


from .turbidostat import TurbidostatCulture
from .culture_functions import increase_stress_drug1
from ipywidgets import interact
import numpy as np


class EnduranceStressCulture(TurbidostatCulture):
    """
    class for subjecting the culture to a constantly increasing stress level (e.g. 5% per generation)
    useful for rapidly increasing the stress (relative to an evolutionary timespan)
    and determining the fitness before significant evolutionary changes occur.

        EXAMPLE:
    culture = EnduranceStressCulture(directory="./NewExperiment/", vial_number=1,
                                     name="Species 1", description="Strain 1", device=None,
                                     default_dilution_volume=10, dead_volume=15, od_max_limit=0.3,
                                     stress_increase_per_generation=1.05, initial_generations=3)
    """
    pumps = (1, 2, 4)

    def __init__(self, directory: str = None, vial_number: int = None, name: str = "Species 1", description: str = "Strain 1",
                 default_dilution_volume: float = 10, dead_volume: float = 15, od_max_limit: float = 0.3,
                 stress_increase_per_generation: float = 1.05, initial_generations: float = 3):
        self.initial_generations = initial_generations
        self.initial_stress = 3
        self.stress_increase_per_generation = stress_increase_per_generation
        super().__init__(directory=directory, vial_number=vial_number, name=name, description=description,
                         default_dilution_volume=default_dilution_volume, dead_volume=dead_volume,
                         od_max_limit=od_max_limit)
        self.pumps = (1, 2, 4)

    def update(self):
        if self.is_active():
            if self.od > self.od_max_limit:
                is_diluting_like_crazy = self.minutes_since_last_dilution < self.minimum_dilution_delay_hrs * 60
                if not is_diluting_like_crazy:
                    if self._log2_dilution_coefficient > self.initial_generations:
                        if self.medium2_concentration < self.initial_stress:
                            self.set_stress(self.initial_stress)
                        else:
                            self.increase_stress()

                    else:
                        self.lower_od()

    def set_stress(self, stress_level):
        increase_stress_drug1(culture=self, target_concentration=stress_level)

    def check(self):
        super(EnduranceStressCulture, self).check()
        dilution_factor = (self.dead_volume + self.default_dilution_volume) / self.dead_volume
        assert self.stress_increase_per_generation < dilution_factor

    def increase_stress(self):
        dilution_factor = (self.dead_volume + self.default_dilution_volume)/self.dead_volume
        generations_per_dilution = np.log2(dilution_factor)
        stress_increase_per_dilution = generations_per_dilution * (self.stress_increase_per_generation - 1) + 1
        increase_stress_drug1(culture=self, stress_increase_factor=stress_increase_per_dilution)

    @staticmethod
    def interactive_demo():
        @interact(doubling_time=(0, 12, 0.1), initial_dose=(0, 0.5, 0.01), factor=(1, 1.5, 0.01), MIC=(0.1, 15, 0.1))
        def interactive_function(t_doubling=3, initial_dose=0.25, factor=1.1, MIC=5):
            generations_per_day = 24 / t_doubling
            factor_per_day = factor ** generations_per_day

            print("         Doubling time: %.1f hours" % t_doubling)
            print("                   MIC: %.1f" % MIC + "% " + "ethanol")
            print("          Initial dose: %.2f of MIC" % initial_dose)
            print("Stress increase factor: %.2f per generation," % factor, "%.3f per day" % factor_per_day)

            # generations_to_mic = np.log(1 / initial_dose) / np.log(factor)
            # print("MIC reached in %.1f generations" % generations_to_mic)
            for ngen in [10, 20, 30, 40, 50]:
                dose_gen = initial_dose * factor ** ngen
                n_days = t_doubling * ngen / 24
                print("    Generation %d" % ngen, "dose: %.2f x MIC" % dose_gen, "(%.1f days)" % n_days)

            print("\nExample:")
            print("Dose at every generation:\n",
                  ", ".join(str(x) for x in np.round(MIC * initial_dose * factor ** np.arange(21), 2)))



import time
import os
import numpy as np
from src.replifactory import read_csv_tail
from src.replifactory import calculate_last_growth_rate
from src.replifactory import log_dilution
from src.replifactory import write_variable
from .blank import BlankCulture
from .culture_functions import inoculate


class TurbidostatCulture(BlankCulture):
    """
    Implements a quasi-turbidostat culture in a multi-vial device.
    A discrete dilution is made when the OD reaches the threshold,
    allowing the pumps to sequentially maintain all 7 vials.
    The growth is more similar to a real, continuous-dilution turbidostat
    when the dilution factor is small and dilutions are made more frequently.
    """
    minimum_dilution_delay_hrs = 0.1
    pumps = (1, 4)

    def __init__(self, directory: str = None, vial_number: int = None, name: str = "Species 1",
                 description: str = "Strain 1", default_dilution_volume: float = 10,
                 dead_volume: float = 15, od_max_limit: float = 0.3):
        # Configuration parameters
        self.default_dilution_volume = default_dilution_volume
        self.dead_volume = dead_volume
        self.od_max_limit = od_max_limit

        # Running parameters
        self._mu = None
        self._mu_error = None
        self._t_doubling = None
        self._t_doubling_error = None
        self._medium2_concentration = 0
        self._medium3_concentration = 0
        self._log2_dilution_coefficient = 0
        self._inoculation_time = None
        self._samples_collected = {}
        self._is_aborted = False
        self._mu_max_measured = 0
        self._t_doubling_min_measured = np.inf
        self._time_last_dilution = {1: None,
                                    2: None,
                                    3: None,
                                    4: None}
        super().__init__(directory=directory, vial_number=vial_number, name=name, description=description)

    def inoculate(self, name=None, description=None):
        inoculate(culture=self, name=name, description=description)
        self._mu_max_measured = 0
        self._t_doubling_min_measured = np.inf
        self._is_aborted = False
        self.save()

    def collect_sample(self, sample_id, sample_volume=1):
        # TODO preserve drug concentrations while sampling
        device = self.device
        vial = self.vial_number
        device.locks_vials[vial].acquire()
        device.lock_pumps.acquire()
        q = input("ready to collect %d ml sample? [y/n]" % sample_volume)
        if q == "y":
            assert not device.is_pumping()  # must have been checked before releasing the pump lock
            assert not device.hard_stop_trigger
            device.stirrers.set_speed(vial=vial, speed=1)
            device.valves.open(vial)  # valve number might be different for e.g. a lagoon setup
            device.pump1.pump(sample_volume)
            device.pump1.stock_volume -= sample_volume
            log_dilution(device=device, vial_number=vial, pump1_volume=sample_volume)
            while device.is_pumping():
                time.sleep(0.5)
            q = input("Is vial volume at vacuum needle level? [y/n]")
            if q != "y":
                vacuum_volume = sample_volume + 5
                device.pump4.pump(vacuum_volume)
                while device.is_pumping():
                    time.sleep(0.5)
                log_dilution(device=device, vial_number=vial, pump4_volume=vacuum_volume)
            self._samples_collected[time.time()] = sample_id
            device.stirrers.set_speed(vial=vial, speed=2)
            assert not device.is_pumping()  # make sure before closing valves
            time.sleep(1)
            device.valves.close(valve=vial)
        device.locks_vials[vial].release()
        device.lock_pumps.release()
        self.device.locks_vials[self.vial_number].acquire()
        self.device.lock_pumps.acquire()

    # def show_parameters(self, increase_verbosity=False):
    #     if self.is_active():
    #         active = "ACTIVE"
    #     else:
    #         active = "NOT ACTIVE"
    #     print("****** Vial %d:" % self.vial_number, active)
    #     if self.is_active() or increase_verbosity:
    #         print("             name:", self.name)
    #         print("      description:", self.description)
    #         print("               od:", np.round(self.od, 4))
    #         print("     od_max_limit:", self.od_max_limit)
    #         print("    last dilution:", np.round(time.time() - self.time_last_dilution, 1)/60, "minutes ago")
    #
    #         print("               mu:", np.round(self.mu, 5), "±", np.round(np.float32(self._mu_error), 5),
    #               "[1/h];     max:", self._mu_max_measured)
    #         print("       t_doubling:", np.round(self.t_doubling, 3), "±", np.round(np.float32(self._t_doubling_error), 3),
    #               "hrs;     min:", self._t_doubling_min_measured)
    #         print("    Generation nr:", self.log2_dilution_coefficient)
    #         if increase_verbosity:
    #             print()
    #             print("      dead_volume:", self.dead_volume)
    #             print("  dilution_volume:", self.default_dilution_volume)
    #             if self._inoculation_time:
    #                 print("       inoculated:", time.ctime(self._inoculation_time))
    #             else:
    #                 print("       inoculated: NO")
    #
    #             print("samples_collected:", self._samples_collected)
    #             print("       is_aborted:", self._is_aborted)
    #             print("         od_blank:", self._od_blank)
    #
    #             vial_volume = self.dead_volume
    #             added_volume = self.default_dilution_volume
    #             dilution_factor = (vial_volume + added_volume) / vial_volume
    #             generations_per_dilution = np.log2(dilution_factor)
    #             generations_per_ml = generations_per_dilution / added_volume
    #             print("generations/Liter: %.1f" % (generations_per_ml * 1000))
    #         print()

    @property
    def log2_dilution_coefficient(self):
        return np.float32(self._log2_dilution_coefficient)

    @log2_dilution_coefficient.setter
    def log2_dilution_coefficient(self, value):
        self._log2_dilution_coefficient = value
        write_variable(culture=self, variable_name="log2_dilution_coefficient", value=value)
        self.save()

    @property
    def mu(self):
        """
        growth rate [1/h]
        :return:
        """
        return np.float32(self._mu)

    @mu.setter
    def mu(self, value):
        mu, error = value
        write_variable(culture=self, variable_name="mu", value=mu)
        write_variable(culture=self, variable_name="mu_error", value=error)
        self._mu = mu
        self._mu_error = error
        if mu != 0:
            t_doubling = np.log(2) / mu
            t_doubling_error = t_doubling * error / mu
            self.t_doubling = (t_doubling, t_doubling_error)

            if mu > self._mu_max_measured and error / mu < 0.05:
                self._mu_max_measured = mu
                self._t_doubling_min_measured = t_doubling
        self.save()

    @property
    def t_doubling(self):
        return np.float32(self._t_doubling)

    @t_doubling.setter
    def t_doubling(self, value):
        t_doubling, t_doubling_error = value
        write_variable(culture=self, variable_name="t_doubling", value=t_doubling)
        write_variable(culture=self, variable_name="t_doubling_error", value=t_doubling_error)
        self._t_doubling = t_doubling
        self._t_doubling_error = t_doubling_error
        self.save()

    def update_growth_rate(self):
        """
        reads last od values and calculates growth rate
        :return:
        """
        od_filepath = os.path.join(self.directory, "od.csv")
        df = read_csv_tail(od_filepath, lines=300)
        df = df[df.index >= df.index[-1]-60*60*5]  # cut last 5 hours
        if np.isfinite(self.time_last_dilution):
            df = df[df.index > np.float32(self.time_last_dilution)]
        t = df.index.values
        od = df.values.ravel() - self.od_blank
        od[od <= 0] = 1e-6
        timepoint, mu, error = calculate_last_growth_rate(t, od)
        if np.isfinite(mu):
            self.mu = (mu, error)

    @property
    def time_last_dilution(self):
        pump_dilution_times = np.array(list(self._time_last_dilution.values())).astype(float)
        most_recent_dilution_time = np.nanmax(pump_dilution_times)
        return np.float32(most_recent_dilution_time)

    @property
    def medium2_concentration(self):
        return np.float32(self._medium2_concentration)

    @medium2_concentration.setter
    def medium2_concentration(self, value):
        self._medium2_concentration = value
        write_variable(culture=self, variable_name="medium2_concentration", value=value)
        self.save()

    @property
    def medium3_concentration(self):
        return np.float32(self._medium3_concentration)

    @medium3_concentration.setter
    def medium3_concentration(self, value):
        self._medium3_concentration = value
        write_variable(culture=self, variable_name="medium3_concentration", value=value)
        self.save()

    @property
    def minutes_since_last_dilution(self):
        if not np.isfinite(self.time_last_dilution):
            seconds_since_last_dilution = time.time() - self._inoculation_time
        else:
            seconds_since_last_dilution = time.time() - self.time_last_dilution
        return np.float32(seconds_since_last_dilution/60)

    def update(self):
        """
        called every minute
        """
        if self.is_active():
            diluting_like_crazy = self.minutes_since_last_dilution < self.minimum_dilution_delay_hrs * 60
            if self.od > np.float32(self.od_max_limit):
                if not diluting_like_crazy:
                    self.lower_od()

    def lower_od(self):
        """
        makes a dilution, preserving the drug concentration in the vial
        :return:
        """
        medium2_target_concentration = self.medium2_concentration
        total_volume = self.dead_volume + self.default_dilution_volume

        drug1_total_amount = total_volume * medium2_target_concentration
        drug1_current_amount = self.dead_volume * self.medium2_concentration
        drug1_pumped_amount = drug1_total_amount - drug1_current_amount
        drug1_pumped_volume = drug1_pumped_amount / self.device.pump2.stock_concentration
        drug1_pumped_volume = min(self.default_dilution_volume, max(0, drug1_pumped_volume))
        if drug1_pumped_volume < 0.05:
            drug1_pumped_volume = 0

        drugfree_medium_volume = self.default_dilution_volume - drug1_pumped_volume  # - drug2_pumped_volume
        drugfree_medium_volume = min(self.default_dilution_volume, max(0, drugfree_medium_volume))

        self.dilute(pump1_volume=drugfree_medium_volume,
                    pump2_volume=drug1_pumped_volume)

    def dilute(self, pump1_volume=0.0, pump2_volume=0.0, pump3_volume=0.0, extra_vacuum=5):
        """
        pump the given volumes into the vial,
        pump the total volume + extra_vacuum out of the vial
        extra_vacuum has to be ~>3 to fill the waste tubing with air and prevent cross-contamination.
        """
        if pump2_volume > 0:
            assert self.medium2_concentration >= 0
            assert self.device.pump2.stock_concentration >= 0
        if pump3_volume > 0:
            assert self.medium3_concentration >= 0
            assert self.device.pump3.stock_concentration >= 0
        self.device.make_dilution(vial=self.vial_number,
                                  pump1_volume=pump1_volume,
                                  pump2_volume=pump2_volume,
                                  pump3_volume=pump3_volume,
                                  extra_vacuum=extra_vacuum)
        self.calculate_culture_concentrations_after_dilution(pump1_volume, pump2_volume, pump3_volume)

    # def do_blank_od(self):
    #     self.od_blank = read_csv_tail(os.path.join(self.directory, "od.csv"), 10).values.mean()


    def calculate_culture_concentrations_after_dilution(self, pump1_volume, pump2_volume, pump3_volume):
        """
        pump1_volume: Drug-free medium
        pump2_volume: Drug 1
        pump3_volume: Drug 2
        """

        dilution_volume = sum([pump1_volume, pump2_volume, pump3_volume])
        total_volume = self.dead_volume + dilution_volume
        dilution_coefficient = total_volume / self.dead_volume
        self.log2_dilution_coefficient = self.log2_dilution_coefficient + np.log2(dilution_coefficient)

        if pump2_volume > 0 or self.medium2_concentration > 0:
            medium2_pumped_amount = self.device.pump2.stock_concentration * pump2_volume
            medium2_vial_amount = self.dead_volume * self.medium2_concentration
            medium2_total_amount = medium2_vial_amount + medium2_pumped_amount
            self.medium2_concentration = medium2_total_amount / total_volume

        if pump3_volume > 0 or self.medium3_concentration > 0:
            medium3_pumped_amount = self.device.pump3.stock_concentration * pump3_volume
            medium3_vial_amount = self.dead_volume * self.medium3_concentration
            medium3_total_amount = medium3_vial_amount + medium3_pumped_amount
            self.medium3_concentration = medium3_total_amount / total_volume
        self.save()

    def check(self):
        assert self.device.is_connected()
        assert self.vial_number in [1, 2, 3, 4, 5, 6, 7]
        assert os.path.exists(self.directory)
        assert 0 < self.dead_volume <= 35
        assert 0 < self.default_dilution_volume <= 40-self.dead_volume
        # assert 0 < self.od_max_limit <= 50
        assert callable(self.device.od_sensors[self.vial_number].calibration_function)
        assert callable(self.device.pump1.calibration_function)
        assert callable(self.device.pump4.calibration_function)
        assert -0.3 < self.od_blank < 0.3
        self.device.stirrers.check_calibration(self.vial_number)

    @property
    def experiment_start_time(self):
        filepath = os.path.join(self.directory, "od.csv")
        if not os.path.exists(filepath):
            return np.nan
        with open(filepath) as f:
            time_first_od = np.float32(f.readlines(512)[1].split(",")[0])
        return np.float32(time_first_od)

    def flush_tubing_if_necessary(self):
        if not self._is_aborted:
            pump_volumes = {1: 0, 2: 0}
            tstart = self.experiment_start_time
            tinoc = np.float32(self._inoculation_time)
            for pump_number in self.pumps:
                tdil = np.float32(self._time_last_dilution[pump_number])
                last_pump_time = np.nanmax([tdil, tstart, tinoc])
                if (time.time() - last_pump_time) > self.device.delay_tube_wash_hrs * 3600:
                    pump_volumes[pump_number] = 0.1
            if pump_volumes[1] > 0 or pump_volumes[2] > 0:
                self.dilute(pump1_volume=pump_volumes[1], pump2_volume=pump_volumes[2], pump3_volume=0)


# class BasicCulture:
#     minimum_dilution_delay_hrs = 0.1
#
#     def __init__(self):
#         # self.experiment_directory = None
#         self.vial_number = None
#         # self.vial_directory = None
#         self.directory = None
#         self._device = None
#         self.file_lock = threading.Lock()
#
#         # Variables logged to csv files
#         self._od = None
#         self._mu = None
#         self._mu_error = None
#         self._t_doubling = None
#         self._t_doubling_error = None
#         self._medium2_concentration = 0
#         self._medium3_concentration = 0
#         self._log2_dilution_coefficient = 0
#
#         # User-assigned experiment parameters
#         self.name = "Culture"
#         self.description = "culture initialized on " + time.ctime()
#         self.od_blank = 0
#         self.dead_volume = 15  # volume below vacuum needle
#
#         self.inoculation_time = None
#         self.samples_collected = {}
#         self.is_aborted = False
#
#         # Running parameters
#         self.mu_max_measured = 0
#         self.t_doubling_min_measured = np.inf
#         self.last_dilution_time = {1: None,
#                                    2: None,
#                                    3: None,
#                                    4: None}
#
#         # if vial_number is not None:
#         #     assert experiment_directory is not None
#         #     assert os.path.exists(experiment_directory)
#         # if experiment_directory is not None:
#         #     assert 1 <= vial_number <= 7
#         #     assert not experiment_directory[:-1].endswith("vial_")
#         #     assert os.path.exists(experiment_directory)
#         #     self.vial_directory = os.path.join(experiment_directory, "vial_%d" % vial_number)
#         #     if not os.path.exists(self.vial_directory):
#         #         os.mkdir(self.vial_directory)
#         #         self.save()
#         #         print("Created new vial directory: %s" % self.vial_directory)
#         #     else:
#         #         self.load()
#
#     def inoculate(self, name, description):
#         assert self.inoculation_time is None, "Vial already inoculated"
#         self.name = name
#         self.description = description
#         self.inoculation_time = int(time.time())
#         self.mu_max_measured = 0
#         self.t_doubling_min_measured = np.inf
#         self.is_aborted = False
#         self.save()
#
#     def collect_sample(self, sample_id, sample_volume=1):
#         # TODO preserve drug concentrations while sampling
#         device = self.device
#         vial = self.vial_number
#         device.locks_vials[vial].acquire()
#         device.lock_pumps.acquire()
#         q = input("ready to collect %d ml sample? [y/n]" % sample_volume)
#         if q == "y":
#             assert not device.is_pumping()  # must have been checked before releasing the pump lock
#             assert not device.hard_stop_trigger
#             device.stirrers.set_speed(vial=vial, speed=1)
#             device.valves.open(vial)  # valve number might be different for e.g. a lagoon setup
#             device.pump1.pump(sample_volume)
#             device.pump1.stock_volume -= sample_volume
#             log_dilution(device=device, vial_number=vial, pump1_volume=sample_volume)
#             while device.is_pumping():
#                 time.sleep(0.5)
#             q = input("Is vial volume at vacuum needle level? [y/n]")
#             if q != "y":
#                 vacuum_volume = sample_volume + 5
#                 device.pump4.pump(vacuum_volume)
#                 while device.is_pumping():
#                     time.sleep(0.5)
#                 log_dilution(device=device, vial_number=vial, pump4_volume=vacuum_volume)
#             self.samples_collected[time.time()] = sample_id
#             device.stirrers.set_speed(vial=vial, speed=2)
#             assert not device.is_pumping()  # make sure before closing valves
#             time.sleep(1)
#             device.valves.close(valve=vial)
#         device.locks_vials[vial].release()
#         device.lock_pumps.release()
#         self.device.locks_vials[self.vial_number].acquire()
#         self.device.lock_pumps.acquire()
#
    def show_parameters(self, increase_verbosity=False):
        if self.is_active():
            active = "ACTIVE"
        else:
            active = "NOT ACTIVE"
        print("****** Vial %d:" % self.vial_number, active)
        if self.is_active() or increase_verbosity:
            print("             name:", self.name)
            print("      description:", self.description)
            print("               od:", np.round(self.od, 4))
            print("   medium 2 conc.:", np.round(self.medium2_concentration, 4))

            # print("     od_max_limit:", self.od_max_limit)
            print("    last dilution:", np.round(time.time() - self.time_last_dilution, 1)/60, "minutes ago")

            print("               mu:", np.round(self.mu, 5), "±", np.round(np.float32(self._mu_error), 5),
                  "[1/h];     max:", self._mu_max_measured)
            print("       t_doubling:", np.round(self.t_doubling, 3), "±", np.round(np.float32(self._t_doubling_error), 3),
                  "hrs;     min:", self._t_doubling_min_measured)
            print("    Generation nr:", self.log2_dilution_coefficient)
            if increase_verbosity:
                print()
                print("      dead_volume:", self.dead_volume)
                print("  dilution_volume:", self.default_dilution_volume)
                if self._inoculation_time:
                    print("       inoculated:", time.ctime(self._inoculation_time))
                else:
                    print("       inoculated: NO")

                print("samples_collected:", self._samples_collected)
                print("       is_aborted:", self._is_aborted)
                print("         od_blank:", self.od_blank)

                vial_volume = self.dead_volume
                added_volume = self.default_dilution_volume
                dilution_factor = (vial_volume + added_volume) / vial_volume
                generations_per_dilution = np.log2(dilution_factor)
                generations_per_ml = generations_per_dilution / added_volume
                print("generations/Liter: %.1f" % (generations_per_ml * 1000))
            print()
#
#     @property
#     def od_deblanked(self):
#         return np.float32(self.od - self.od_blank)
#
#     @property
#     def od(self):
#         return np.float32(self._od)
#
#     @od.setter
#     def od(self, value):
#         write_variable(culture=self, variable_name="od", value=value)
#         self._od = value
#         self.update_growth_rate()
#         self.save()
#
#     @property
#     def log2_dilution_coefficient(self):
#         return np.float32(self._log2_dilution_coefficient)
#
#     @log2_dilution_coefficient.setter
#     def log2_dilution_coefficient(self, value):
#         self._log2_dilution_coefficient = value
#         write_variable(culture=self, variable_name="log2_dilution_coefficient", value=value)
#         self.save()
#
#     @property
#     def mu(self):
#         """
#         growth rate [1/h]
#         :return:
#         """
#         return np.float32(self._mu)
#
#     @mu.setter
#     def mu(self, value):
#         mu, error = value
#         write_variable(culture=self, variable_name="mu", value=mu)
#         write_variable(culture=self, variable_name="mu_error", value=error)
#         self._mu = mu
#         self._mu_error = error
#         if mu != 0:
#             t_doubling = np.log(2) / mu
#             t_doubling_error = t_doubling * error / mu
#             self.t_doubling = (t_doubling, t_doubling_error)
#
#             if mu > self.mu_max_measured and error / mu < 0.05:
#                 self.mu_max_measured = mu
#                 self.t_doubling_min_measured = t_doubling
#         self.save()
#
#     @property
#     def t_doubling(self):
#         return np.float32(self._t_doubling)
#
#     @t_doubling.setter
#     def t_doubling(self, value):
#         t_doubling, t_doubling_error = value
#         write_variable(culture=self, variable_name="t_doubling", value=t_doubling)
#         write_variable(culture=self, variable_name="t_doubling_error", value=t_doubling_error)
#         self._t_doubling = t_doubling
#         self._t_doubling_error = t_doubling_error
#         self.save()
#
#     def update_growth_rate(self):
#         """
#         reads last od values and calculates growth rate
#         :return:
#         """
#         od_filepath = os.path.join(self.directory, "od.csv")
#         df = read_csv_tail(od_filepath, lines=300)
#         df = df[df.index >= df.index[-1]-60*60*5]  # cut last 5 hours
#         if np.isfinite(self.time_last_dilution):
#             df = df[df.index > np.float32(self.time_last_dilution)]
#         t = df.index.values
#         od = df.values.ravel() - self.od_blank
#         od[od <= 0] = 1e-6
#         timepoint, mu, error = calculate_last_growth_rate(t, od)
#         if np.isfinite(mu):
#             self.mu = (mu, error)
#
#     @property
#     def time_last_dilution(self):
#         pump_dilution_times = np.array(list(self.last_dilution_time.values())).astype(float)
#         most_recent_dilution_time = np.nanmax(pump_dilution_times)
#         return np.float32(most_recent_dilution_time)
#         # filepath = os.path.join(self.vial_directory, "dilutions.csv")
#         # if not os.path.exists(filepath):
#         #     return np.nan
#         # data = read_csv_tail(filepath=filepath, lines=1)
#         # last_timepoint = data.index.ravel()[-1]
#         # return last_timepoint
#
#     @property
#     def medium2_concentration(self):
#         return np.float32(self._medium2_concentration)
#
#     @medium2_concentration.setter
#     def medium2_concentration(self, value):
#         self._medium2_concentration = value
#         write_variable(culture=self, variable_name="medium2_concentration", value=value)
#         self.save()
#
#     @property
#     def medium3_concentration(self):
#         return np.float32(self._medium3_concentration)
#
#     @medium3_concentration.setter
#     def medium3_concentration(self, value):
#         self._medium3_concentration = value
#         write_variable(culture=self, variable_name="medium3_concentration", value=value)
#         self.save()
#
#     @property
#     def minutes_since_last_dilution(self):
#         if not np.isfinite(self.time_last_dilution):
#             seconds_since_last_dilution = time.time() - self.inoculation_time
#         else:
#             seconds_since_last_dilution = time.time() - self.time_last_dilution
#         return np.float32(seconds_since_last_dilution/60)
#
#     def plot(self, last_hours=24):
#         fig = plot_culture(culture=self, last_hours=last_hours)
#         return fig
#
#     def is_active(self):
#         return np.isfinite(np.float32(self.inoculation_time)) and not self.is_aborted
#
#     def save(self):
#         """
#         saves config to vial_directory/culture_config.yaml
#         """
#         self.file_lock.acquire()  # maybe not necessary?
#         config_path = os.path.join(self.directory, "culture_config.yaml")
#         save_object(self, filepath=config_path)
#         self.file_lock.release()
#
#     def load(self):
#         config_path = os.path.join(self.directory, "culture_config.yaml")
#         self.file_lock.acquire()
#         load_config(self, filepath=config_path)
#         self.file_lock.release()
#         print("Loaded vial %d config." % self.vial_number)
#
#     @property
#     def device(self):
#         return self._device
#
#     @device.setter
#     def device(self, dev):
#         if dev is not None:
#             if self.directory is not None:
#                 assert os.path.join(dev.directory, "vial_%d" % self.vial_number) == self.directory
#         self._device = dev
#         self.save()
#
#     def update(self):
#         """
#         called every minute
#         """
#         pass
#
#     def dilute(self, pump1_volume=0.0, pump2_volume=0.0, pump3_volume=0.0, extra_vacuum=5):
#         """
#         pump the given volumes into the vial,
#         pump the total volume + extra_vacuum out of the vial
#         extra_vacuum has to be ~>3 to fill the waste tubing with air and prevent cross-contamination.
#         """
#         if pump2_volume > 0:
#             assert self.medium2_concentration >= 0
#             assert self.device.pump2.stock_concentration >= 0
#         if pump3_volume > 0:
#             assert self.medium3_concentration >= 0
#             assert self.device.pump3.stock_concentration >= 0
#         self.device.make_dilution(vial=self.vial_number,
#                                   pump1_volume=pump1_volume,
#                                   pump2_volume=pump2_volume,
#                                   pump3_volume=pump3_volume,
#                                   extra_vacuum=extra_vacuum)
#         self.calculate_culture_concentrations_after_dilution(pump1_volume, pump2_volume, pump3_volume)
#
#     def do_blank_od(self):
#         od_blank = read_csv_tail(os.path.join(self.directory, "od.csv"), 10).values.mean()
#         self.od_blank = od_blank
#
#     def calculate_culture_concentrations_after_dilution(self, pump1_volume, pump2_volume, pump3_volume):
#         """
#         pump1_volume: Drug-free medium
#         pump2_volume: Drug 1
#         pump3_volume: Drug 2
#         """
#
#         dilution_volume = sum([pump1_volume, pump2_volume, pump3_volume])
#         total_volume = self.dead_volume + dilution_volume
#         dilution_coefficient = total_volume / self.dead_volume
#         self.log2_dilution_coefficient = self.log2_dilution_coefficient + np.log2(dilution_coefficient)
#
#         if pump2_volume > 0 or self.medium2_concentration > 0:
#             medium2_pumped_amount = self.device.pump2.stock_concentration * pump2_volume
#             medium2_vial_amount = self.dead_volume * self.medium2_concentration
#             medium2_total_amount = medium2_vial_amount + medium2_pumped_amount
#             self.medium2_concentration = medium2_total_amount / total_volume
#
#         if pump3_volume > 0 or self.medium3_concentration > 0:
#             medium3_pumped_amount = self.device.pump3.stock_concentration * pump3_volume
#             medium3_vial_amount = self.dead_volume * self.medium3_concentration
#             medium3_total_amount = medium3_vial_amount + medium3_pumped_amount
#             self.medium3_concentration = medium3_total_amount / total_volume
#         self.save()
#
#     def check(self):
#         assert self.device.is_connected()
#         assert 1 <= self.vial_number <= 7
#         assert os.path.exists(self.directory)
#         assert 0 < self.dead_volume <= 35
#         assert 0 < self.default_dilution_volume <= 40-self.dead_volume
#         assert 0 < self.od_max_limit <= 50
#         assert callable(self.device.od_sensors[self.vial_number].calibration_function)
#         assert callable(self.device.pump1.calibration_function)
#         assert callable(self.device.pump4.calibration_function)
#         assert -0.3 < self.od_blank < 0.3
#         self.device.stirrers.check_calibration(self.vial_number)
#
#     @property
#     def experiment_start_time(self):
#         filepath = os.path.join(self.directory, "od.csv")
#         if not os.path.exists(filepath):
#             return np.nan
#         with open(filepath) as f:
#             time_first_od = np.float32(f.readlines(512)[1].split(",")[0])
#         return time_first_od
#
#     def flush_tubing_if_necessary(self):
#         if not self.is_aborted:
#             pump_volumes = {1: 0, 2: 0}
#             tstart = self.experiment_start_time
#             tinoc = np.float32(self.inoculation_time)
#             for pump_number in [1, 2]:
#                 tdil = self.last_dilution_time[pump_number]
#                 last_pump_time = np.nanmax(np.array([tdil, tstart, tinoc]).astype(np.float32))
#                 if (time.time() - last_pump_time) * 3600 < self.device.delay_tube_wash_hrs:
#                     pump_volumes[pump_number] = 0.1
#             if pump_volumes[1] > 0 or pump_volumes[2] > 0:
#                 self.dilute(pump1_volume=pump_volumes[1], pump2_volume=pump_volumes[2], pump3_volume=0)



import time
import threading
import pyftdi.i2c


class PwmController:
    """PCA9685 PWM controller"""
    def __init__(self, device, frequency=50):
        self.device = device
        self.frequency = frequency
        self.lock = threading.Lock()  # valves and stirrers can be used on different threads
        self.port = None
        if self.device.is_connected():
            self.connect()

    def connect(self):
        try:
            self.port = self.device.i2c.get_port(self.device.PORT_PWM)
            self.set_frequency(self.frequency)
            for led_number in range(16):
                led_on_l = led_number * 4 + 6
                led_on_h = led_number * 4 + 7
                self.port.write_to(led_on_h, [0x0])
                self.port.write_to(led_on_l, [0x00])
            self.start_all()
        except pyftdi.i2c.I2cNackError:
            self.port = None
            print("PCA9685 PWM controller connection ERROR.")

    def set_frequency(self, frequency):
        pre_scale = round(25000000 / (4096 * frequency)) - 1
        try:
            self.port.write_to(0x00, [0b00010001])  # sleep mode
        except:
            time.sleep(0.5)
            self.port.write_to(0x00, [0b00010001])  # sleep mode
            self.port.write_to(0x00, [0b0])  # reset
            print("Reset PWM driver")
            self.port.write_to(0x00, [0b00010001])  # sleep mode
        self.port.write_to(0xFE, [pre_scale])  # SET_PWM_FREQUENCY
        self.port.write_to(0x00, [0b10000001])  # restart mode

    def get_duty_cycle(self, led_number):
        pass

    def set_duty_cycle(self, led_number, duty_cycle):
        assert 0 <= led_number <= 15
        assert 0 <= duty_cycle <= 1
        msb, lsb = divmod(round(4095 * duty_cycle), 0x100)  # most and least significant bytes
        led_off_l = led_number * 4 + 8
        led_off_h = led_number * 4 + 9
        self.port.write_to(led_off_l, [lsb])
        self.port.write_to(led_off_h, [msb])

    def stop_all(self):
        self.port.write_to(0x00, [0b10001])

    def start_all(self):
        self.port.write_to(0x00, [0b00001])
        time.sleep(0.002)
        self.port.write_to(0x00, [0b10000001])

    def is_sleeping(self):
        mode1_register = self.port.read_from(0x00, 1)[0]
        is_sleeping = bool(int(bin(mode1_register)[2:].rjust(8, '0')[-5]))  # sleep bit
        return is_sleeping



import ipywidgets as widgets
from IPython.core.display import clear_output, display
from ipywidgets import VBox, HBox, Layout


class CulturePlotWidget:
    def __init__(self, main_gui, vial_number):
        self.main_gui = main_gui
        self.title = widgets.HTML('<b>Vial %d:</b>' % vial_number, layout=Layout(width="40px", height="40px"))
        self.vial_number = vial_number
        self.last_hours = widgets.IntText(description="time window",
                                          description_tooltip="hours before the last OD measurement", value=24,
                                          style={'description_width': '80px'}, layout=Layout(width="135px"))
        self.plot_growth_rate = widgets.Checkbox(description="T_doubling", value=False, indent=False,
                                                 layout=Layout(width="135px"))
        self.output = widgets.Output(layout=Layout(width="800px"))
        self.button = widgets.Button(description="plot")
        self.button.on_click(self.handle_plot_button_click)
        config = VBox([self.title, self.button, self.last_hours, self.plot_growth_rate], layout=Layout(width="155px"))
        self.widget = HBox([config, self.output])

    def handle_plot_button_click(self, b):
        self.button.disabled = True
        self.button.description = "plotting..."
        try:
            with self.output:
                c = self.main_gui.experiment.device.cultures[self.vial_number]
                fig = c.plot(last_hours=self.last_hours.value, plot_growth_rate=self.plot_growth_rate.value)
                clear_output()
                display(fig)
        finally:
            self.button.description = "plot"
            self.button.disabled = False


class ExperimentPlotTab:
    title = "Experiment plot"

    def __init__(self, main_gui):
        self.main_gui = main_gui
        self.culture_plots = [CulturePlotWidget(main_gui=main_gui, vial_number=v) for v in range(1,8)]
        self.plots = [culture_plot.widget for culture_plot in self.culture_plots]
        self.plot_all = widgets.Button(description="plot all vials")
        self.plot_all.on_click(self.handle_plot_all_button)
        self.widget = VBox([self.plot_all, VBox(self.plots)])

    def handle_plot_all_button(self, button):
        button.disabled = True
        for culture_plot in self.culture_plots:
            culture_plot.button.click()
        button.disabled = False


import ipywidgets as widgets
from IPython.core.display import clear_output
from ipywidgets import interactive, fixed, HBox, Layout, VBox
from src.replifactory import increase_stress_drug1

widget_layout = Layout(align_items='center', width='90px')
box_layout = Layout(display='flex',
                    flex_flow='row',
                    align_items='center',
                    border='solid',
                    width='720px')
style = {'description_width': '0px'}


class StirrerWidget:
    def __init__(self, device, vial_number):
        self.device = device
        self.vial_number = vial_number
        self.slider = widgets.FloatSlider(0, min=0, max=1, step=0.01,
                                          orientation="vertical",
                                          description="Mix %d" % vial_number,
                                          continuous_update=False,
                                          layout=Layout(width='30px'))
        self.slider.observe(self.handle_slider_change, names="value")
        self.max_button = widgets.Button(description="max", layout=Layout(width='60px'))
        self.max_button.on_click(self.handle_max_button)
        self.high_button = widgets.Button(description="high", layout=Layout(width='60px'))
        self.high_button.on_click(self.handle_high_button)
        self.low_button = widgets.Button(description="low", layout=Layout(width='60px'))
        self.low_button.on_click(self.handle_low_button)
        self.stop_button = widgets.Button(description="STOP", layout=Layout(width='60px'))
        self.stop_button.on_click(self.handle_stop_button)
        self.calibrate = widgets.Checkbox(description="learn", layout=Layout(width='60px'),
                                          style={"description_width": "initial"})
        self.buttons = widgets.VBox([self.max_button, self.high_button, self.low_button,
                                     self.stop_button, self.calibrate])
        self.widget = HBox([self.slider, self.buttons], layout=Layout(align_items='center', width='100px'))

    def handle_stop_button(self, button):
        self.slider.value = 0

    def handle_max_button(self, button):
        if self.calibrate.value:
            self.device.calibration_fan_speed_to_duty_cycle[self.vial_number][3] = self.slider.value
            self.calibrate.value = False
        else:
            self.slider.value = self.device.calibration_fan_speed_to_duty_cycle[self.vial_number][3]

    def handle_high_button(self, button):
        if self.calibrate.value:
            self.device.calibration_fan_speed_to_duty_cycle[self.vial_number][2] = self.slider.value
            self.calibrate.value = False
        else:
            self.slider.value = self.device.calibration_fan_speed_to_duty_cycle[self.vial_number][2]

    def handle_low_button(self, button):
        if self.calibrate.value:
            self.device.calibration_fan_speed_to_duty_cycle[self.vial_number][1] = self.slider.value
            self.calibrate.value = False
        else:
            self.slider.value = self.device.calibration_fan_speed_to_duty_cycle[self.vial_number][1]

    def handle_slider_change(self, change):
        new_speed = change.new
        self.device.stirrers._set_duty_cycle(vial=self.vial_number, duty_cycle=new_speed)


class StirrerWidgets:
    def __init__(self, device):
        self.device = device
        self.max_all = widgets.Button(description="max RPM")
        self.max_all.on_click(self.handle_all_max_button)

        self.high_all = widgets.Button(description="high RPM")
        self.high_all.on_click(self.handle_all_high_button)

        self.low_all = widgets.Button(description="low RPM")
        self.low_all.on_click(self.handle_all_low_button)

        self.stop_all = widgets.Button(description="STOP stirrers", button_style="danger")
        self.stop_all.on_click(self.handle_all_stop_button)

        self.control_all = HBox([self.max_all, self.high_all, self.low_all, self.stop_all])
        self.stirrer_widgets = [StirrerWidget(self.device, i) for i in range(1, 8)]
        stirrer_widgets_box = HBox([w.widget for w in self.stirrer_widgets])
        self.widget = VBox([self.control_all, stirrer_widgets_box],layout=Layout(display='flex',
                    flex_flow='column',
                    align_items='center',
                    border='solid',
                    width='720px'))

    def handle_all_max_button(self, button):
        for w in self.stirrer_widgets:
            w.calibrate.value = False
            w.max_button.click()

    def handle_all_high_button(self, button):
        for w in self.stirrer_widgets:
            w.calibrate.value = False
            w.high_button.click()

    def handle_all_low_button(self, button):
        for w in self.stirrer_widgets:
            w.calibrate.value = False
            w.low_button.click()

    def handle_all_stop_button(self, button):
        for w in self.stirrer_widgets:
            w.stop_button.click()


class ValveButtons:
    def __init__(self, device):
        self.layout = Layout(align_items='center', width='90px', height="30px")
        self.device = device
        self.valve_buttons = []
        self.open_all = widgets.Button(description="OPEN all valves")
        self.close_all = widgets.Button(description="CLOSE all valves")
        self.open_all.on_click(self.handle_open_all_button)
        self.close_all.on_click(self.handle_close_all_button)
        for v in range(1, 8):
            button = widgets.ToggleButton(value=self.device.valves.is_open[v],
                                          description="Valve %d" % v, layout=self.layout, style=style)
            button.observe(handler=self.handle_valve_button, names=["value", "description"])

            self.valve_buttons += [button]
        box_valves = HBox(self.valve_buttons, layout=Layout(display='flex',
                                                            flex_flow='row',
                                                            align_items='stretch',
                                                            # height='120px',
                                                            width='720px'))
        all_valves = HBox([self.open_all, self.close_all])
        self.widget = VBox([box_valves, all_valves], layout=Layout(display='flex',
                                                                   flex_flow='column',
                                                                   align_items='stretch',
                                                                   border='solid',
                                                                   width='720px'))

    def handle_valve_button(self, change):
        valve = int(change.owner.description[-1])
        if self.device.locks_vials[valve].locked():
            print("ERROR: LOCKED")
        else:
            self.device.locks_vials[valve].acquire()
            self.device.valves.set_state(valve=valve, is_open=change.owner.value)
            # if is_open:
            #     print("Open")
            # else:
            #     print("Closed")
            self.device.locks_vials[valve].release()

    #
    # def make_valve_button(self, valve_number):
    #     button = widgets.ToggleButton(value=is_open, description="Valve %d" % valve_number,
    #                                   layout=self.layout, style=style)
    #
    #     is_open = self.device.valves.is_open[valve_number]
    #     button = widgets.ToggleButton(value=is_open, description="Valve %d" % valve_number,
    #                                   layout=self.layout, style=style)
    #     widget = interactive(self.valve_button, valve=fixed(valve_number), is_open=button, layout=self.layout)
    #     return widget

    def handle_open_all_button(self, b):
        self.open_all.disabled = True
        # self.device.valves.open_all()
        for v in range(1, 8):
            if not self.device.valves.is_open[v] is True:
                self.valve_buttons[v-1].value = True
        self.close_all.disabled = False

    def handle_close_all_button(self, b):
        self.close_all.disabled = True
        # self.device.valves.close_all()
        for v in range(1, 8):
            if not self.device.valves.is_open[v] is False:
                self.valve_buttons[v-1].value = False
        self.open_all.disabled = False

# def valve_button(self, valve, is_open=False):
#     if self.device.locks_vials[valve].locked():
#         print("ERROR: LOCKED")
#     else:
#         self.device.locks_vials[valve].acquire()
#         self.device.valves.set_state(valve=valve, is_open=is_open)
#         # if is_open:
#         #     print("Open")
#         # else:
#         #     print("Closed")
#         self.device.locks_vials[valve].release()



class DilutionWidget:
    def __init__(self, device):
        self.device = device
        self.vial = widgets.Dropdown(options=[1, 2, 3, 4, 5, 6, 7], index=None, description="Vial")
        self.target_concentration = widgets.FloatSlider(disabled=True, description="concentration")
        self.vial.observe(self.handle_vial_change, names="value")
        self.button = widgets.Button(description="make dilution")
        self.button.on_click(self.handle_button)
        self.widget = VBox([self.vial, self.target_concentration, self.button])

    def handle_button(self, button):
        self.button.disabled = True
        self.button.description = "diluting..."
        c = self.device.cultures[self.vial.value]
        increase_stress_drug1(culture=c, target_concentration=self.target_concentration.value)
        self.button.disabled = False
        self.button.description = "make dilution"

    def handle_vial_change(self, change):
        c = self.device.cultures[change["new"]]
        dilution_factor = (c.default_dilution_volume + c.dead_volume) / c.dead_volume
        self.target_concentration.min = c.medium2_concentration / dilution_factor
        stock_c = c.device.pump2.stock_concentration
        self.target_concentration.value = c.medium2_concentration
        self.target_concentration.max = (
                                                    c.medium2_concentration * c.dead_volume + c.default_dilution_volume * stock_c) / (
                                                    c.dead_volume + c.default_dilution_volume)
        self.target_concentration.disabled = False


class DeviceControl:
    def __init__(self, device):
        self.device = device
        if self.device is None:
            self.widget = widgets.Output()
            with self.widget:
                print("No device connected.")
        else:
            self.widget = self.build_widget(self.device)

    def build_widget(self, device):
        # STIRRERS ##
        # stirrer_widgets = []
        # for stirrer in range(1, 8):
        #     slider = widgets.FloatSlider(0, min=0, max=1, step=0.01,
        #                                  orientation="vertical",
        #                                  description="Stirrer %d" % stirrer,
        #                                  continuous_update=False,
        #                                  layout=widget_layout)
        #     stirrer_slider = interactive(device.stirrers._set_duty_cycle, vial=fixed(stirrer), duty_cycle=slider)
        #     stirrer_widgets += [stirrer_slider]
        # box_stirrers = HBox(stirrer_widgets, layout=box_layout

        box_stirrers = StirrerWidgets(device).widget
    # LASERS ####
        laser_widgets = []

        def laser_trigger(vial, is_on=False):
            if is_on:
                device.lasers.switch_on(vial)
            else:
                device.lasers.switch_off(vial)

        for laser in range(1, 8):
            checkbox = widgets.Checkbox(value=False, description="Laser %d" % laser,
                                        layout=widget_layout, style=style)
            laser_checkbox = interactive(laser_trigger, vial=fixed(laser), is_on=checkbox)
            laser_widgets += [laser_checkbox]
        box_lasers = HBox(laser_widgets, layout=box_layout)

    # ADC #######
        adc_widgets = []
        outputs = {v: widgets.Output() for v in range(1, 8)}

        def make_adc_button_function(photodiode):
            def button_function(b):
                with outputs[photodiode]:
                    self.device.photodiodes.switch_to_vial(vial=photodiode)
                    clear_output()
                    mv, err = device.photodiodes.measure(gain=8, bitrate=16)
                    print(mv, "\n±%.7f" % err)
            return button_function

        for photodiode in range(1, 8):
            button = widgets.Button(description='ADC %d' % photodiode, height="20px", layout=Layout(width='90px'))
            button.on_click(make_adc_button_function(photodiode))
            sub_box = VBox([button, outputs[photodiode]],
                           layout=Layout(align_items='center', width='110px', height="70px"))
            adc_widgets += [sub_box]
        box_adc = HBox(adc_widgets, layout=box_layout)

    # OD ########
        od_widgets = []
        od_outputs = {v: widgets.Output() for v in range(1, 8)}

        def make_od_button_function(vial):
            def button_function(b):
                with od_outputs[vial]:
                    clear_output()
                    od = device.od_sensors[vial].measure_od()
                    try:
                        device.cultures[vial].od = od
                    except:
                        pass
                    print("%.4f" % od)
            return button_function

        for vial in range(1, 8):
            button = widgets.Button(description='OD %d' % vial, height="20px", layout=Layout(width='90px'))
            button.on_click(make_od_button_function(vial))
            sub_box = VBox([button, od_outputs[vial]],
                           layout=Layout(align_items='center', width='110px', height="70px"))
            od_widgets += [sub_box]
        box_od = HBox(od_widgets, layout=box_layout)

    # PUMPS #####
        pump_style = {'description_width': 'initial'}
        rotations_sliders = {pump: widgets.FloatSlider(0, min=0, max=40, step=0.001, orientation="horizontal",
                                                       description="PUMP %d rotations" % pump, continuous_update=False,
                                                       style=pump_style) for pump in range(1, 5)}
        run_buttons = {pump: widgets.Button(description="RUN pump %d" % pump) for pump in range(1, 5)}

        def run_pump1(b):
            device.pump1.move(n_rotations=rotations_sliders[1].value)

        def run_pump2(b):
            device.pump2.move(n_rotations=rotations_sliders[2].value)

        def run_pump3(b):
            device.pump3.move(n_rotations=rotations_sliders[3].value)

        def run_pump4(b):
            device.pump4.move(n_rotations=rotations_sliders[4].value)

        run_buttons[1].on_click(run_pump1)
        run_buttons[2].on_click(run_pump2)
        run_buttons[3].on_click(run_pump3)
        run_buttons[4].on_click(run_pump4)

        box_pumps = []
        for pump in range(1, 5):
            box_pumps += [HBox([rotations_sliders[pump], run_buttons[pump]])]

        def emergency_stop_button_action(b):
            self.device.pump1.stop()
            self.device.pump2.stop()
            self.device.pump3.stop()
            self.device.pump4.stop()
            # self.device.emergency_stop()

        def continuous_vacuum_button_action(b):
            assert self.device.valves.not_all_closed()
            self.device.pump4.run(0.1)

        emergency_stop = widgets.Button(description="STOP pumps", button_style="danger")
        continuous_vacuum = widgets.Button(description="continuous vacuum", button_style="warning")
        continuous_vacuum.on_click(continuous_vacuum_button_action)
        emergency_stop.on_click(emergency_stop_button_action)

        box_pumps += [HBox([continuous_vacuum, emergency_stop])]
        box_pumps += [DilutionWidget(self.device).widget]
        pumps_box_layout = Layout(display='flex',
                            flex_flow='column',
                            align_items='stretch',
                            border='solid',
                            width='470px')
        box_pumps = VBox(box_pumps, layout=pumps_box_layout)
    # TEMP ######
        temp_output = widgets.Output()

        def temp_measure(b):
            with temp_output:
                clear_output()
                cold, hot = device.thermometers.measure_temperature()
                print("Vials:\t%.2f °C\nBoard:\t%.2f °C" % (cold, hot))

        temp_button = widgets.Button(description="Temperature")
        temp_button.on_click(temp_measure)
        box_temps = VBox([temp_button, temp_output], layout=Layout(height="80px"))

        box_pumps_and_temps = HBox([box_pumps, box_temps])
    # TAB #######
        widget_control_whole = VBox([box_stirrers,
                                     box_lasers,
                                     ValveButtons(self.device).widget,
                                     box_adc,
                                     box_od,
                                     box_pumps_and_temps], disabled=True)
        return widget_control_whole


import inspect
import ipywidgets as widgets
import pyftdi.ftdi
from pyftdi.usbtools import UsbTools
from ipywidgets import VBox, HBox
from src.replifactory import BaseDevice, MorbidostatDevice, TurbidostatCulture, MorbidostatCulture
from src.replifactory import DeviceControl
import glob
import os
from IPython.display import clear_output


class DeviceTab:

    def __init__(self, main_gui):
        self.title = "Device control"
        config_paths = glob.glob("../**/device_config.yaml", recursive=True)
        self.config_paths = [os.path.relpath(p) for p in config_paths]
        self.experiment_directories = [os.path.relpath(os.path.join(p, "..")) for p in self.config_paths]

        self.main_gui = main_gui
        self.status_bar = main_gui.status_bar
        self.device_class = widgets.Dropdown(description="*CLASS", options=get_device_classes(), index=None)
        self.ftdi_address = widgets.Dropdown(description="*ADDRESS", options=get_ftdi_addresses(), index=None)
        self.config_path = widgets.Dropdown(description="CALIBRATION", options=self.config_paths, index=None)
        self.connect_button = widgets.Button(description="connect device")
        self.reset_connection_button = widgets.Button(description="reset connections", button_style="danger")
        self.reset_connection_button.on_click(self.handle_connection_reset_button)
        self.connect_button.on_click(self.handle_connect_button)
        self.control_widget = DeviceControl(None).widget
        self.widget = None
        self.update()

    def update(self):
        self.widget = VBox([self.device_class,
                            self.ftdi_address,
                            self.config_path,
                            HBox([self.connect_button, self.reset_connection_button]),
                            self.control_widget])

    def update_selection_options(self):
        config_paths = glob.glob("../**/device_config.yaml", recursive=True)
        self.config_paths = [os.path.relpath(p) for p in config_paths]
        self.experiment_directories = [os.path.relpath(os.path.join(p, "..")) for p in self.config_paths]
        self.device_class = widgets.Dropdown(description="*CLASS", options=get_device_classes(), index=None)
        self.ftdi_address = widgets.Dropdown(description="*ADDRESS", options=get_ftdi_addresses(), index=None)
        self.config_path = widgets.Dropdown(description="CALIBRATION", options=self.config_paths, index=None)
        self.update()
        self.main_gui.update()

    def handle_connect_button(self, button):
        button.disabled = True
        button.description = "connecting..."
        try:
            self.connect_device()
        finally:
            button.disabled = False
            button.description = "reconnect device"

    def handle_connection_reset_button(self, button):
        UsbTools.release_all_devices()
        UsbTools.flush_cache()
        # self.control_widget = DeviceControl(None).widget
        self.update_selection_options()

    def connect_device(self):
        self.control_widget = DeviceControl(None).widget
        with self.status_bar.output:
            clear_output()
            self.update()
            self.main_gui.device = self.device_class.value(ftdi_address=self.ftdi_address.value)
            self.main_gui.device.connect()
            if self.config_path.value is not None:
                self.main_gui.device.load_calibration(config_path=self.config_path.value)
            self.control_widget = DeviceControl(self.main_gui.device).widget
            self.update()
        self.main_gui.update()

    # def connect_experiment_device(self):
    #     with self.status_bar.output:
    #         clear_output()
    #         self.widget = DeviceControl(self.main_gui.device).widget
    #     self.main_gui.update()


def get_device_classes():
    ks = list(globals().keys())
    device_classes = [BaseDevice, MorbidostatDevice]
    for k in ks:
        try:
            if BaseDevice in inspect.getmro(globals()[k]):
                device_classes += [globals()[k]]
        except:
            pass
    culture_classes = list(set(device_classes))
    return culture_classes


def get_culture_classes():
    ks = list(globals().keys())
    culture_classes = [TurbidostatCulture, MorbidostatCulture]
    for k in ks:
        try:
            if TurbidostatCulture in inspect.getmro(globals()[k]):
                culture_classes += [globals()[k]]
        except:
            pass
    culture_classes = list(set(culture_classes))
    return culture_classes


def get_ftdi_addresses():
    available_devices = [d[0] for d in pyftdi.ftdi.Ftdi.list_devices()]
    ftdi_addresses = []
    for value in available_devices:
        bus = value.bus
        address = value.address
        ftdi_addresses += ["ftdi://ftdi:2232h:%d:%d" % (bus, address)]
    return ftdi_addresses


from src.replifactory import read_csv_tail
import numpy as np
from scipy.optimize import curve_fit
import pandas as pd
import matplotlib.pyplot as plt


def growth_function(t, N0, growth_rate):
    y = N0 * np.exp(growth_rate * t)
    return y


def calculate_growth_rate(time_values, od_values):
    """
    fits exponential growth curve and returns growth rate calculated over the entire time window.
    time window must not include dilution.
    time values must be in seconds.
    """
    assert 1 < time_values[-1] < 1e10  # must be time in seconds since 1970-01-01 00:00:00

    if len(time_values) < 6:  # at least 6 data points required for reasonable measurement.
        return np.nan, np.nan, np.nan

    if np.diff(time_values).max() > 30*60:  # half an hour gap in the data, major problem.
        return np.nan, np.nan, np.nan

    # denoise, trim
    if len(time_values) >= 10:
        median_window = int(len(time_values) / 8)+1
        od_values = pd.Series(od_values).rolling(median_window * 2, center=True).median()
        time_values = time_values[median_window:-(median_window - 1)]
        od_values = od_values[median_window:-(median_window - 1)]

    # convert to hrs
    hours = np.array(time_values) / 3600
    hours = hours - min(hours)

    try:
        popt, pcov = curve_fit(growth_function, hours, od_values, p0=(1e-3, 0.3))
    except:
        return np.nan, np.nan, np.nan
    timepoint = (time_values[0] + time_values[-1]) / 2
    growth_rate = popt[1]
    growth_rate_error = np.sqrt(np.diag(pcov))[1]
    return timepoint, growth_rate, growth_rate_error


def get_last_gr(od_filepath):
    """
    returns the last growth rate in an od.csv file,
    calculated over an appropriate time window
    """
    df = read_csv_tail(od_filepath, lines=300)
    df = df[df.index >= df.index[-1]-60*60*5]  # cut last 5 hours
    t = df.index.values
    od = df.values.ravel()
    timepoint, growth_rate, error = calculate_last_growth_rate(t, od)
    return timepoint, growth_rate, error


def calculate_last_growth_rate(t, od):
    """
    returns the most recent growth rate in an OD sequence,
    calculated over an appropriate time window (~t_doubling/2)
    """
    if len(od) == 0:
        return np.nan, np.nan, np.nan
    od[od <= 0] = 1e-6
    od_delta = abs(od.max() - od.min())
    min_window_size = 10
    if od_delta < 0.1:
        max_window_size = 300
    else:
        max_window_size = 60

    timepoint, growth_rate, error = np.nan, np.nan, np.nan
    # initial guess
    window_size = min(60, (t[-1]-t[0])/60)  # minutes
    guessed_td = np.nan
    tmax = t[-1]

    for _ in range(4):
        tmin = tmax - window_size * 60
        imin = np.where(t >= tmin)[0][0]
        tw = t[imin:]
        odw = od[imin:]
        timepoint, growth_rate, error = calculate_growth_rate(tw, odw)
        td = np.log(2) / growth_rate  # hours
        if not np.isfinite(td):
            break
        guess_error = abs((td - guessed_td) / td)
        guessed_td = td

        new_window_size = int(abs(td) * 60 * 0.5)  # optimum window size [minutes] is ~half the doubling time
        new_window_size = max(new_window_size, min_window_size)
        new_window_size = min(new_window_size, max_window_size)
        if guess_error < 0.05:
            break
        if t[0] > tmin:
            break
        if new_window_size == window_size:
            break
        else:
            window_size = new_window_size

    return timepoint, growth_rate, error


# def sliding_window_growth_rate(time_values, od_values, window_size_minutes):
#     """
#     time_values in seconds
#     window size in minutes
#     """
#     time_values = np.array(time_values)
#     od_values = np.array(od_values)
#     od_values[od_values <= 0] = 1e-6
#
#     growth_rates = []
#     errors = []
#     timepoints = []
#     for i in range(len(time_values)):
#         tmin = time_values[i]
#         tmax = tmin + window_size_minutes * 60
#         if tmax <= time_values[-1]:
#             imax = np.where(time_values >= tmax)[0][0]
#             time_window = time_values[i:imax + 1]
#             od_window = od_values[i:imax + 1]
#
#             timepoint = (tmin+tmax)/2
#             try:
#                 growth_rate, error = calculate_growth_rate(time_window, od_window)
#             except:
#                 growth_rate, error = np.nan, np.nan
#             timepoints += [timepoint]
#             growth_rates += [growth_rate]
#             errors += [error]
#     growth_rates = np.array(growth_rates)
#     errors = np.array(errors)
#     timepoints = np.array(timepoints)
#     return timepoints, growth_rates, errors
#
#
# def sliding_window_doubling_time(t, od, window_size_minutes=20):
#     timepoints, growth_rates, gr_errors = sliding_window_growth_rate(t, od, window_size_minutes=window_size_minutes)
#     t_doubling = np.log(2) / growth_rates
#     t_doubling_error = t_doubling * gr_errors / growth_rates
#     return timepoints, t_doubling, t_doubling_error

    # for i in range(len(od)):
    #     imax = i + 10
    #     if len(od) <= imax:
    #         tmax = t[imax]
    #         imax = np.where(t >= tmax)[0][0]
    #         tw = t[i:imax + 1]
    #         odw = od[i:imax + 1]
    #         timepoint, growth_rate, error = calculate_last_growth_rate(tw, odw)
    #         timepoints += [timepoint]
    #         growth_rates += [growth_rate]
    #         errors += [error]


def adaptive_window_growth_rate(t, od, dilution_timepoints=None):
    timepoints, growth_rates, errors = [], [], []
    for i in range(len(od)-10):
        jmin = i
        jmax = jmin+10
        if t[jmax] - t[0] <= 3600 * 5:
            jmin = 0
        else:
            tmax = t[jmax]
            tmin = tmax - 3600 * 5
            if dilution_timepoints is not None:
                tdil = [t for t in dilution_timepoints if tmin < t < tmax]
                if len(tdil) > 0:
                    tmin = tdil[-1]
            jmin = np.where(t >= tmin)[0][0]

        tw = t[jmin:jmax]
        odw = od[jmin:jmax]

        timepoint, growth_rate, error = calculate_last_growth_rate(tw, odw)
        timepoints += [timepoint]
        growth_rates += [growth_rate]
        errors += [error]

    growth_rates = np.array(growth_rates)
    timepoints = np.array(timepoints)
    errors = np.array(errors)
    return timepoints, growth_rates, errors


def adaptive_window_doubling_time(t, od, dilution_timepoints):
    timepoints, growth_rates, gr_errors = adaptive_window_growth_rate(t, od, dilution_timepoints)
    t_doubling = np.log(2) / growth_rates
    t_doubling_error = t_doubling * gr_errors / growth_rates
    return timepoints, t_doubling, t_doubling_error


def plot_gr(time_values, od_values, dilution_timepoints=None):
    od = np.array(od_values)
    t = np.array(time_values)

    fig, ax = plt.subplots(figsize=[16, 8], dpi=100);
    # i = 0
    # lines = []
    ax.plot(t / 3600, od, "k.", label="Optical Density")
    # ax.set_ylim(-0.05, 1.6)
    od[od <= 0] = 1e-6
    ax.set_ylabel("Optical Density")
    ax.set_xlabel("Time [hours]")
    ax2 = ax.twinx()

    # if not adaptive_window:
    #     colors = ["xkcd:dark red", "xkcd:light orange", "xkcd:green", "xkcd:purple"]
    #     for c, ws in zip(colors[:len(window_sizes_minutes)], window_sizes_minutes):
    #         td_timepoints, td, tderr = sliding_window_doubling_time(t, od, window_size_minutes=ws)
    #
    #         markers, caps, bars = ax2.errorbar(td_timepoints / 3600, td, tderr, color=c,
    #                                            alpha=0.5, label="doubling time, %d min time windows" % ws)
    #         [bar.set_alpha(0.1) for bar in bars]

    td_timepoints, td, tderr = adaptive_window_doubling_time(t, od, dilution_timepoints=dilution_timepoints)

    markers, caps, bars = ax2.errorbar(td_timepoints / 3600, td, tderr,
                                       alpha=0.5, label="doubling time")
    [bar.set_alpha(0.1) for bar in bars]
    try:
        td = np.array(td)
        tderr = np.array(tderr)
        tdmax = np.nanmax(td[tderr < 0.05])
        tdmin = np.nanmin(td[tderr < 0.05])
        ax2.set_ylim(tdmin*0.5, tdmax*1.2)
    except:
        pass

    ax2.grid()
    ax2.set_ylabel("Doubling time [hours]")
    # labels = [line.get_label() for line in lines]
    # ax.legend(loc=1)
    # ax2.legend(loc=2)
    # plt.show()
    return fig





from .blank import BlankCulture
import numpy as np
import os
from .culture_functions import inoculate


class BatchCulture(BlankCulture):

    """
    class for subjecting the culture to a constantly increasing stress level (e.g. 5% per generation)
    useful for rapidly increasing the stress (relative to an evolutionary timespan)
    and determining the fitness before significant evolutionary changes occur.

        EXAMPLE:
    culture = EnduranceStressCulture(directory="./NewExperiment/", vial_number=1,
                                     name="Species 1", description="Strain 1", device=None,
                                     default_dilution_volume=10, dead_volume=15, od_max_limit=0.3,
                                     stress_increase_per_generation=1.05, initial_generations=3)
    """
    pumps = []

    def __init__(self, directory: str = None, vial_number: int = None, name: str = "Species 1",
                 description: str = "Strain 1"):
        # Running parameters
        self._mu = None
        self._mu_error = None
        self._t_doubling = None
        self._t_doubling_error = None
        self._inoculation_time = None
        self._samples_collected = {}
        self._is_aborted = False
        self._mu_max_measured = 0
        self._t_doubling_min_measured = np.inf
        super().__init__(directory=directory, vial_number=vial_number, name=name, description=description)
        del self.dead_volume

    def update(self):
        pass

    def check(self):
        assert self.device.is_connected()
        assert self.vial_number in [1, 2, 3, 4, 5, 6, 7]
        assert os.path.exists(self.directory)
        self.device.stirrers.check_calibration(self.vial_number)
        assert callable(self.device.od_sensors[self.vial_number].calibration_function)
        assert -0.3 < self.od_blank < 0.3

    def inoculate(self, name=None, description=None):
        inoculate(culture=self, name=name, description=description)
        self._mu_max_measured = 0
        self._t_doubling_min_measured = np.inf
        self._is_aborted = False
        self.save()



import os
import yaml
import time
import pickle
import numpy as np
import pyftdi
import inspect


def load_config(obj, filepath):
    config = open(filepath).read()
    loaded_dict = yaml.load(config, Loader=yaml.Loader)
    assert loaded_dict["_class"] == obj.__class__
    for k in loaded_dict.keys():
        if k != "directory":
            obj.__dict__[k] = loaded_dict[k]


def load_object(filepath):
    """
    reads a python object from a yaml config
    """
    if not os.path.exists(filepath):
        return None
    else:
        config = open(filepath).read()
        loaded_dict = yaml.load(config, Loader=yaml.Loader)
        klass = loaded_dict["_class"]
        s = inspect.signature(klass.__init__)
        init_args = list(s.parameters.keys())
        init_args = [arg for arg in init_args if arg != "self"]
        kwargs = {arg: loaded_dict[arg] for arg in init_args if arg in loaded_dict.keys()}
        obj = klass(**kwargs)
        for k in loaded_dict.keys():
            obj.__dict__[k] = loaded_dict[k]

        sep = os.path.sep
        if sep not in filepath:
            sep = os.path.altsep
        assert sep in filepath, "Directory must contain '\\' or '/'"

        directory = os.path.join(*filepath.split(sep)[:-1])
        obj.directory = directory
        return obj


def clean_variable(var):
    if type(var) is dict:
        return {k: clean_variable(var[k]) for k in var.keys()}
    try:
        if np.isfinite(var):
            if type(var) is int:
                return int(var)
            else:
                return float(var)
    except:
        return var


def save_object(obj, filepath):
    d = obj.__dict__.copy()
    d["_class"] = obj.__class__
    d["_saved_on"] = time.ctime()
    for k in list(d.keys()):
        try:
            pickle.dumps(d[k])
            d[k] = clean_variable(d[k])
        except:
            d.pop(k)
    with open(filepath, "w+") as f:
        f.write(yaml.dump(d))


def save_experiment(exp):
    if exp.device is not None:
        exp.device.save()
    for culture in exp.cultures.values():
        if culture is not None:
            culture.save()


def load_experiment(exp):
    directory = exp.directory
    if not os.path.exists(directory):
        os.mkdir(directory)
        print("Created new experiment directory: %s" % directory)
    else:
        print("Experiment directory: '%s'" % directory)
        t0 = time.time()
        device_config_path = os.path.join(directory, "device_config.yaml")
        if os.path.exists(device_config_path):
            exp.device = load_object(filepath=device_config_path)
            exp.device.load()  # load cultures
        else:
            print("No device config found.")



import time
import numpy as np
import scipy.optimize
from scipy.optimize import fsolve
import matplotlib.pyplot as plt
from src.replifactory import Stepper


def calibrate_pump(device, pump_number):
    p = device.pumps[pump_number]
    n_rotations = float(input("How many rotations?"))
    n_repetitions = float(input("How many repetitions?"))
    assert n_repetitions % 1 == 0
    n_repetitions = int(n_repetitions)
    total_volume = n_rotations * n_repetitions * 0.1
    p.fit_calibration_function()
    if callable(p.calibration_function):
        def opt_function(volume):
            return p.calibration_function(volume) - n_rotations

        predicted_mls = fsolve(opt_function, 1)[0]
        predicted_total_mls = predicted_mls * n_repetitions
        print("Predicted %.2f mls from existing calibration data" % predicted_total_mls)
        total_volume = predicted_total_mls
    q = input("Are ~%d ml available? \nZero the scale and press ENTER; q to exit" % round(total_volume))
    if q == "":
        for i in range(n_repetitions):
            p.move(n_rotations)
            print("%d/%d" % (i + 1, n_repetitions), end="\t\r")
            while p.is_pumping():
                time.sleep(0.1)
            time.sleep(1)
        ml = float(input("How many mls?"))
        ml = ml / n_repetitions
        device.calibration_pump_rotations_to_ml[pump_number][n_rotations] = round(ml, 3)
        device.save()


def pump_calibration_function(x, a, b, c):
    """
    converts pumped volume (x) to motor rotations(y)
    """
    y = a * x + b * x * np.exp(-c * x)
    return y


class Pump(Stepper):
    def __init__(self, device, cs):
        super().__init__(device=device, cs=cs)

        self.device = device
        self.cs = cs
        self.pump_number = cs + 1
        self.calibration_function = None
        if len(self.calibration_rotations_to_ml.keys()) >= 4:
            self.fit_calibration_function()

    def show_parameters(self):
        print("pump %d stock volume:" % self.pump_number, self.stock_volume)
        print("pump %d stock concentration:" % self.pump_number, self.stock_concentration)
        self.calibration_curve()
        plt.show()

    @property
    def stock_concentration(self):
        return np.float32(self.device.pump_stock_concentrations[self.pump_number])

    @stock_concentration.setter
    def stock_concentration(self, value):
        value = float(value)
        self.device.pump_stock_concentrations[self.pump_number] = value
        self.device.save()

    @property
    def stock_volume(self):
        return np.float32(self.device.pump_stock_volumes[self.pump_number])

    @stock_volume.setter
    def stock_volume(self, value):
        value = float(value)
        self.device.pump_stock_volumes[self.pump_number] = value
        self.device.save()

    @property
    def calibration_rotations_to_ml(self):
        return self.device.calibration_pump_rotations_to_ml[self.pump_number]

    @calibration_rotations_to_ml.setter
    def calibration_rotations_to_ml(self, value):
        self.device.calibration_pump_rotations_to_ml[self.pump_number] = value
        self.fit_calibration_function()
        self.device.save()

    def pump(self, volume):
        rotations = self.calibration_function(abs(volume))
        if volume < 0:
            rotations *= -1
        if volume != 0:
            self.move(n_rotations=rotations)

    def add_calibration_point(self, rotations, volume):
        temp = self.calibration_rotations_to_ml
        temp[rotations] = volume
        self.calibration_rotations_to_ml = temp

    def calibrate(self, rotations_to_ml=None, dummy_data=False):
        self.device.save()
        self.calibration_rotations_to_ml = rotations_to_ml
        if rotations_to_ml is None and dummy_data:
            self.calibration_rotations_to_ml = {1: 0.145,
                                                5: 0.563,
                                                10: 1.012,
                                                20: 2.14,
                                                50: 5.14,
                                                100: 9.21}

    def fit_calibration_function(self):
        calibration_n_rotations = sorted(list(self.calibration_rotations_to_ml.keys()))
        calibration_volumes = sorted(list(self.calibration_rotations_to_ml.values()))
        if calibration_n_rotations is not None:
            if len(calibration_n_rotations) >= 4:
                coefs, _ = scipy.optimize.curve_fit(pump_calibration_function, calibration_volumes,
                                                    calibration_n_rotations,
                                                    p0=(10, -10, 1),
                                                    bounds=[(-50, -50, 0.1), (50, 50, 1)],
                                                    sigma=np.array(calibration_volumes) * 0.1)
                a, b, c = coefs

                def f(volume):
                    """
                    converts volume to number of pump rotations
                    """
                    return pump_calibration_function(volume, a, b, c)

                self.calibration_function = f
                return a, b, c

    def calibration_curve(self):

        calibration_n_rotations = sorted(list(self.calibration_rotations_to_ml.keys()))
        calibration_volumes = sorted(list(self.calibration_rotations_to_ml.values()))
        if len(calibration_n_rotations) < 4:
            print("PUMP %d: NOT ENOUGH CALIBRATION POINTS (%d/4)" % (self.pump_number, len(calibration_n_rotations)))
        else:
            a, b, c = self.fit_calibration_function()
            plt.figure(figsize=[6, 4], dpi=100)
            x = np.linspace(0, 20, 101)
            y = [self.calibration_function(vol) for vol in x]
            plt.plot(calibration_volumes, calibration_n_rotations, "bs")
            plt.plot(x, y, "k:")
            plt.xlabel("pumped volume [mL]")
            plt.ylabel("motor rotations")
            plt.title("Pump %d calibration\n a (slope): %.3f    b: %.3f    c: %.3f" % (self.pump_number, a, b, c))
            # plt.show()



from .culture_functions import increase_stress_drug1
from .turbidostat import TurbidostatCulture
import time
import numpy as np
import os
import threading


class CsvDataLogger:
    def __init__(self, directory, variable_name):
        self.lock = threading.Lock()
        header = "time,%s\n" % variable_name
        if not header.endswith("\n"):
            header = header+"\n"
        self.header = header
        self.filepath = os.path.join(directory, variable_name+".csv")
        if not os.path.exists(self.filepath):
            with open(self.filepath, "w+") as f:
                f.write(header)

    def log_line(self, line):
        if not line.endswith("\n"):
            line = line+"\n"
        self.lock.acquire()
        try:
            with open(self.filepath, "a") as f:
                f.write(line)
        finally:
            self.lock.release()


class MorbidostatCulture(TurbidostatCulture):
    def __init__(self, directory: str = None, vial_number: int = None, name: str = "Species 1",
                 description: str = "Strain 1", default_dilution_volume: float = 10, dead_volume: float = 15,
                 od_max_limit: float = 0.3):
        self.t_doubling_max_limit = 24
        self.t_doubling_min_limit = 4
        self.delay_stress_increase_hrs = 4
        self.delay_rescue_dilution_hrs = 16
        self.default_dilution_volume = 10
        self.od_max_limit = 0.3

        # Running parameters
        self._last_stress_increase_time = None
        self._mu_max_measured = 0
        self._t_doubling_min_measured = np.inf
        super().__init__(directory=directory, vial_number=vial_number, name=name, description=description,
                         default_dilution_volume=default_dilution_volume, dead_volume=dead_volume,
                         od_max_limit=od_max_limit)

    def update(self):
        """
        called every minute
        """
        if self.is_active():
            self.update_growth_rate()
            growing_too_fast = self.t_doubling < self.t_doubling_min_limit
            time_to_increase_stress = self.hrs_since_stress_increase > self.delay_stress_increase_hrs
            diluting_like_crazy = self.minutes_since_last_dilution < self.minimum_dilution_delay_hrs * 60

            if self.od > np.float32(self.od_max_limit):
                if not diluting_like_crazy:
                    if not growing_too_fast and time_to_increase_stress:
                        self.increase_stress()
                    else:
                        self.lower_od()
            else:
                self.rescue_if_necessary()

    @property
    def hrs_since_stress_increase(self):
        if not np.isfinite(np.float32(self._last_stress_increase_time)):
            return np.float32(time.time() - self._inoculation_time) / 3600
        else:
            return (time.time() - self._last_stress_increase_time) / 3600

    def rescue_if_necessary(self):
        hrs_inactive = self.minutes_since_last_dilution / 60
        if hrs_inactive > np.float32(self.delay_rescue_dilution_hrs):
            if self.t_doubling > np.float32(self.t_doubling_max_limit) or self.t_doubling < 0:
                self.decrease_stress()

    # def increase_stress(self, stress_increase_factor=None):
    #     dilution_factor = (self.dead_volume + self.default_dilution_volume) / self.dead_volume
    #     total_volume = self.dead_volume + self.default_dilution_volume
    #     if stress_increase_factor is None:
    #         stress_increase_factor = (dilution_factor + 1) / 2
    #     medium2_target_concentration = self.medium2_concentration * stress_increase_factor
    #     # medium3_target_concentration = None
    #
    #     drug1_total_amount = total_volume * medium2_target_concentration
    #     drug1_current_amount = self.dead_volume * self.medium2_concentration
    #     drug1_pumped_amount = drug1_total_amount - drug1_current_amount
    #     drug1_pumped_volume = drug1_pumped_amount / self.device.pump2.stock_concentration
    #     drug1_pumped_volume = round(drug1_pumped_volume, 3)
    #     drug1_pumped_volume = min(self.default_dilution_volume, max(0.1, drug1_pumped_volume))
    #     # drug2_pumped_volume = 0
    #     # if medium3_target_concentration:
    #     #     drug2_total_amount = total_volume * medium3_target_concentration
    #     #     drug2_current_amount = self.dead_volume * self.medium3_concentration
    #     #     drug2_pumped_amount = drug2_total_amount - drug2_current_amount
    #     #     drug2_pumped_volume = drug2_pumped_amount / self.device.pump3.stock_concentration
    #     #     drug2_pumped_volume = min(self.default_dilution_volume, max(0, drug2_pumped_volume))
    #
    #     drugfree_medium_volume = self.default_dilution_volume - drug1_pumped_volume   # - drug2_pumped_volume
    #     drugfree_medium_volume = min(self.default_dilution_volume, max(0, drugfree_medium_volume))
    #
    #     self.dilute(pump1_volume=drugfree_medium_volume,
    #                 pump2_volume=drug1_pumped_volume)
    #     self._last_stress_increase_time = int(time.time())



    def increase_stress(self, stress_increase_factor=None):
        increase_stress_drug1(culture=self, stress_increase_factor=stress_increase_factor)
        self._last_stress_increase_time = int(time.time())

    def decrease_stress(self):
        """
        makes standard dilution with no drug
        :return:
        """
        # dilution_factor = (self.dead_volume + self.default_dilution_volume) / self.dead_volume
        # stress_decrease_factor = dilution_factor
        self.dilute(pump1_volume=self.default_dilution_volume,
                    pump2_volume=0)

    def check(self):
        super().check()
        assert np.isfinite(self.t_doubling_min_limit)
        assert np.isfinite(self.t_doubling_max_limit)
        assert np.isfinite(self.delay_rescue_dilution_hrs)
        assert np.isfinite(self.delay_stress_increase_hrs)
        assert np.isfinite(self.device.pump_stock_concentrations[2])
        assert callable(self.device.pump2.calibration_function)

# class MorbidostatCulture(BasicCulture):
#     def __init__(self):
#         # User-assigned experiment parameters
#         self.t_doubling_max_limit = 24
#         self.t_doubling_min_limit = 4
#         self.delay_stress_increase_hrs = 4
#         self.delay_rescue_dilution_hrs = 16
#         self.default_dilution_volume = 10
#         self.od_max_limit = 0.3
#         # Running parameters
#         self.last_stress_increase_time = None
#         self.mu_max_measured = 0
#         self.t_doubling_min_measured = np.inf
#
#         super().__init__()
#
#     def show_parameters(self, increase_verbosity=False):
#         super().show_parameters(increase_verbosity=increase_verbosity)
#         if self.is_active() or increase_verbosity:
#             print("medium2_concentration:", self.medium2_concentration, "mM")
#             # print("medium 3 concentration", self.medium3_concentration)
#             print(" t_doubling_max_limit:", self.t_doubling_max_limit)
#             print(" t_doubling_min_limit:", self.t_doubling_min_limit)
#             print("stress increase delay: %.1f hours after last stress increase" % np.float32(self.delay_stress_increase_hrs))
#             print("         rescue delay: %.1f hours after last dilution" % np.float32(self.delay_rescue_dilution_hrs))
#             print()
#
#     # def update(self):
#     #     """
#     #     called every minute
#     #     """
#     #
#     #     if self.is_active():
#     #         growing_too_fast = self.t_doubling < self.t_doubling_min_limit
#     #         time_to_increase_stress = self.hrs_since_stress_increase > self.delay_stress_increase_hrs
#     #         diluting_like_crazy = self.minutes_since_last_dilution < self.minimum_dilution_delay_hrs * 60
#     #
#     #         if self.od_deblanked > np.float32(self.od_max_limit):
#     #             if not diluting_like_crazy:
#     #                 if growing_too_fast and time_to_increase_stress:
#     #                     self.increase_stress()
#     #                 else:
#     #                     self.lower_od()
#     #         else:
#     #             self.rescue_if_necessary()
#     #
#     # @property
#     # def hrs_since_stress_increase(self):
#     #     if not np.isfinite(np.float32(self.last_stress_increase_time)):
#     #         return np.float32(time.time() - self.inoculation_time) / 3600
#     #     else:
#     #         return (time.time() - self.last_stress_increase_time) / 3600
#     #
#     # def rescue_if_necessary(self):
#     #     hrs_inactive = self.minutes_since_last_dilution / 60
#     #     if hrs_inactive > np.float32(self.delay_rescue_dilution_hrs):
#     #         if self.t_doubling > np.float32(self.t_doubling_max_limit) or self.t_doubling < 0:
#     #             self.decrease_stress()
#     #
#     # def increase_stress(self, stress_increase_factor=None):
#     #     dilution_factor = (self.dead_volume + self.default_dilution_volume) / self.dead_volume
#     #     total_volume = self.dead_volume + self.default_dilution_volume
#     #     if stress_increase_factor is None:
#     #         stress_increase_factor = (dilution_factor + 1) / 2
#     #     medium2_target_concentration = self.medium2_concentration * stress_increase_factor
#     #     # medium3_target_concentration = None
#     #
#     #     drug1_total_amount = total_volume * medium2_target_concentration
#     #     drug1_current_amount = self.dead_volume * self.medium2_concentration
#     #     drug1_pumped_amount = drug1_total_amount - drug1_current_amount
#     #     drug1_pumped_volume = drug1_pumped_amount / self.device.pump2.stock_concentration
#     #     drug1_pumped_volume = round(drug1_pumped_volume, 3)
#     #     drug1_pumped_volume = min(self.default_dilution_volume, max(0.1, drug1_pumped_volume))
#     #     # drug2_pumped_volume = 0
#     #     # if medium3_target_concentration:
#     #     #     drug2_total_amount = total_volume * medium3_target_concentration
#     #     #     drug2_current_amount = self.dead_volume * self.medium3_concentration
#     #     #     drug2_pumped_amount = drug2_total_amount - drug2_current_amount
#     #     #     drug2_pumped_volume = drug2_pumped_amount / self.device.pump3.stock_concentration
#     #     #     drug2_pumped_volume = min(self.default_dilution_volume, max(0, drug2_pumped_volume))
#     #
#     #     drugfree_medium_volume = self.default_dilution_volume - drug1_pumped_volume   # - drug2_pumped_volume
#     #     drugfree_medium_volume = min(self.default_dilution_volume, max(0, drugfree_medium_volume))
#     #
#     #     self.dilute(pump1_volume=drugfree_medium_volume,
#     #                 pump2_volume=drug1_pumped_volume)
#     #     self.last_stress_increase_time = int(time.time())
#     #
#     # def lower_od(self):
#     #     """
#     #     makes a dilution, preserving the drug concentration in the vial
#     #     :return:
#     #     """
#     #     medium2_target_concentration = self.medium2_concentration
#     #     total_volume = self.dead_volume + self.default_dilution_volume
#     #
#     #     drug1_total_amount = total_volume * medium2_target_concentration
#     #     drug1_current_amount = self.dead_volume * self.medium2_concentration
#     #     drug1_pumped_amount = drug1_total_amount - drug1_current_amount
#     #     drug1_pumped_volume = drug1_pumped_amount / self.device.pump2.stock_concentration
#     #     drug1_pumped_volume = min(self.default_dilution_volume, max(0, drug1_pumped_volume))
#     #     if drug1_pumped_volume < 0.05:
#     #         drug1_pumped_volume = 0
#     #
#     #     drugfree_medium_volume = self.default_dilution_volume - drug1_pumped_volume  # - drug2_pumped_volume
#     #     drugfree_medium_volume = min(self.default_dilution_volume, max(0, drugfree_medium_volume))
#     #
#     #     self.dilute(pump1_volume=drugfree_medium_volume,
#     #                 pump2_volume=drug1_pumped_volume)
#     #
#     # def decrease_stress(self):
#     #     """
#     #     makes standard dilution with no drug
#     #     :return:
#     #     """
#     #     # dilution_factor = (self.dead_volume + self.default_dilution_volume) / self.dead_volume
#     #     # stress_decrease_factor = dilution_factor
#     #     self.dilute(pump1_volume=self.default_dilution_volume,
#     #                 pump2_volume=0)
#     #
#     # def check(self):
#     #     super().check()
#     #     assert np.isfinite(self.t_doubling_min_limit)
#     #     assert np.isfinite(self.t_doubling_max_limit)
#     #     assert np.isfinite(self.delay_rescue_dilution_hrs)
#     #     assert np.isfinite(self.delay_stress_increase_hrs)
#     #     assert np.isfinite(self.device.pump_stock_concentrations[2])
#     #     assert callable(self.device.pump2.calibration_function)
#



import numpy as np
import threading
import os
from src.replifactory import write_variable
from src.replifactory.loading import save_object, load_config
from src.replifactory import plot_culture
import ipywidgets as widgets

from .culture_functions import is_active
from src.replifactory import read_csv_tail


class BlankCulture:
    pumps = (1, 4)

    def __init__(self, directory: str = None, vial_number: int = None, name: str ="Blank",
                 description: str ="control vial, not inoculated"):
        self.vial_number = vial_number
        self.directory = None
        if directory is not None:
            self.directory = os.path.realpath(directory)
            if not self.directory[:-1].endswith("vial_"):
                self.directory = os.path.join(self.directory, "vial_%d" % vial_number)

        self.name = name
        self.description = description

        self.file_lock = threading.Lock()
        self._device = None
        self.dead_volume = 15  # volume below vacuum needle
        self.od_blank = 0
        self._od = None
        self._od_raw = None

    def update(self):
        pass

    @property
    def parameters(self):
        return sorted([k for k in self.__dict__.keys() if not k.startswith("_")])

    @property
    def od(self):
        return np.float32(self._od)

    @od.setter
    def od(self, value):
        write_variable(culture=self, variable_name="od_plus_blank", value=value)
        self._od_raw = value
        self._od = np.float32(value - self.od_blank)
        write_variable(culture=self, variable_name="od", value=self._od)
        self.save()

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, dev):
        if dev is not None:
            if self.directory is not None:
                assert os.path.join(dev.directory, "vial_%d" % self.vial_number) == self.directory
        self._device = dev
        self.save()

    def save(self):
        """
        saves config to vial_directory/culture_config.yaml
        """
        self.file_lock.acquire()  # maybe not necessary?
        if not os.path.exists(self.directory):
            os.mkdir(self.directory)
        config_path = os.path.join(self.directory, "culture_config.yaml")

        save_object(self, filepath=config_path)
        self.file_lock.release()

    def write_blank_od(self):
        od_plus_blank_filepath = os.path.join(self.directory, "od_plus_blank.csv")
        df = read_csv_tail(filepath=od_plus_blank_filepath, lines=5)
        self.od_blank = df.values.ravel().mean()
        self.save()

    def load(self):
        config_path = os.path.join(self.directory, "culture_config.yaml")
        self.file_lock.acquire()
        try:
            load_config(self, filepath=config_path)
            print("Loaded vial %d config." % self.vial_number)
        finally:
            self.file_lock.release()

    def plot(self, last_hours=24, plot_growth_rate=False):
        fig = plot_culture(culture=self, last_hours=last_hours, plot_growth_rate=plot_growth_rate)
        return fig

    def is_active(self):
        if type(self) is BlankCulture:
            return False
        else:
            return is_active(culture=self)

    def show_parameters(self, increase_verbosity=False):
        if increase_verbosity:
            keys = [k for k in self.__dict__.keys() if not k.startswith("_")]
            keys = [k for k in keys if k not in ["directory", "file_lock"]]
            for k in sorted(keys):
                print("%s:" % k, self.__dict__[k])
        else:
            print("Vial %d: BLANK culture" % self.vial_number)

    def handle_value_change(self, change):
        parameter_name = change.owner.description
        self.__dict__[parameter_name] = change.new
        self.save()

    def check(self):
        assert self.device.is_connected()
        assert self.vial_number in [1, 2, 3, 4, 5, 6, 7]
        assert os.path.exists(self.directory)
        assert 0 < self.dead_volume <= 35
        assert callable(self.device.od_sensors[self.vial_number].calibration_function)
        assert callable(self.device.pump1.calibration_function)
        assert callable(self.device.pump4.calibration_function)
        assert -0.3 < self.od_blank < 0.3
        self.device.stirrers.check_calibration(self.vial_number)

    @property
    def widget(self):
        user_parameters = [k for k in self.__dict__.keys() if not k.startswith("_") and
                           k not in ["name", "description", "directory", "file_lock", "vial_number", "pumps"]]
        parameter_style = {'description_width': '230px'}
        parameter_widgets = [widgets.FloatText(value=self.__dict__[par], description=par, style=parameter_style,
                                               continuous_update=True) for par in user_parameters]
        for w in parameter_widgets:
            w.observe(self.handle_value_change, names="value")
        parameter_box = widgets.VBox(parameter_widgets)

        # description_style = {}
        # description_widgets = [widgets.HTML('<b>Vial %d:</b>' % self.vial_number, layout=Layout(width="40px")),
        #                        widgets.Text(value=self.name, description="name", style=description_style, continuous_update=True),
        #                        widgets.Textarea(value=self.description, description="description",
        #                                         style=description_style, continuous_update=True)]
        # for w in description_widgets:
        #     w.observe(self.handle_value_change, names="value")
        # description_box = widgets.VBox(description_widgets)
        # widgets.HBox([description_box, parameter_box])
        return parameter_box


import pyftdi.i2c


def set_bit(v, index, x):  # source: stackoverflow.com/questions/12173774/
    """Set the index:th bit of v to 1 if x is truthy, else to 0, and return the new value."""
    mask = 1 << index   # Compute mask, an integer with just bit 'index' set.
    v &= ~mask          # Clear the bit indicated by the mask (if x is False)
    if x:
        v |= mask       # If x was True, set the bit indicated by the mask.
    return v            # Return the result, we're done.


class Lasers:
    registers = {1: 2,
                 2: 2,
                 3: 2,
                 4: 2,
                 5: 3,
                 6: 3,
                 7: 3}
    bits = {1: 1,
            2: 3,
            3: 5,
            4: 7,
            5: 1,
            6: 3,
            7: 5}

    def __init__(self, device):
        self.device = device
        self.multiplexer_port = None

        if self.device.is_connected():
            self.connect()

    def connect(self):
        try:
            self.multiplexer_port = self.device.i2c.get_port(self.device.PORT_GPIO_MULTIPLEXER_LASERS)
            self.multiplexer_port.write_to(6, [0x00])  # set all GPIO pins as output
            self.multiplexer_port.write_to(7, [0x00])  # set all GPIO pins as output
        except pyftdi.i2c.I2cNackError:
            print("PCA9555 laser multiplexer connection ERROR.")

    def switch_on(self, vial):
        register = self.registers[vial]
        bit = self.bits[vial]
        old_byte = self.multiplexer_port.read_from(register, 1)[0]  # read which lasers are currently shining
        new_byte = set_bit(old_byte, bit, 0)
        self.multiplexer_port.write_to(register, [new_byte])

    def switch_off(self, vial):
        register = self.registers[vial]
        bit = self.bits[vial]
        old_byte = self.multiplexer_port.read_from(register, 1)[0]  # read which lasers are currently shining
        new_byte = set_bit(old_byte, bit, 1)
        self.multiplexer_port.write_to(register, [new_byte])

    # def lasers_off(self):
    #     self.multiplexer_port.write_to(2, [0xff])  # all 3.3V
    #     self.multiplexer_port.write_to(3, [0xff])  # all 3.3V



import time
import numpy as np


def inoculate(culture, name=None, description=None):
    assert culture._inoculation_time is None, "Vial already inoculated"
    if name is not None:
        culture.name = name
    if description is not None:
        culture.description = description
    culture._inoculation_time = int(time.time())


def is_active(culture):
    return np.isfinite(np.float32(culture._inoculation_time)) and not culture._is_aborted


def increase_stress_drug1(culture, stress_increase_factor=None, target_concentration=None):
    dilution_factor = (culture.dead_volume + culture.default_dilution_volume) / culture.dead_volume
    total_volume = culture.dead_volume + culture.default_dilution_volume
    if target_concentration is None:
        if stress_increase_factor is None:

            stress_increase_factor = (dilution_factor + 1) / 2
        medium2_target_concentration = culture.medium2_concentration * stress_increase_factor
        if culture.medium2_concentration == 0:
            medium2_target_concentration = culture.device.pump2.stock_concentration / 50
    else:
        medium2_target_concentration = target_concentration

    # medium3_target_concentration = None

    drug1_total_amount = total_volume * medium2_target_concentration
    drug1_current_amount = culture.dead_volume * culture.medium2_concentration
    drug1_pumped_amount = drug1_total_amount - drug1_current_amount
    drug1_pumped_volume = drug1_pumped_amount / culture.device.pump2.stock_concentration
    drug1_pumped_volume = round(drug1_pumped_volume, 3)
    drug1_pumped_volume = min(culture.default_dilution_volume, max(0.1, drug1_pumped_volume))
    if target_concentration == 0:
        drug1_pumped_volume = 0
    # drug2_pumped_volume = 0
    # if medium3_target_concentration:
    #     drug2_total_amount = total_volume * medium3_target_concentration
    #     drug2_current_amount = self.dead_volume * self.medium3_concentration
    #     drug2_pumped_amount = drug2_total_amount - drug2_current_amount
    #     drug2_pumped_volume = drug2_pumped_amount / self.device.pump3.stock_concentration
    #     drug2_pumped_volume = min(self.default_dilution_volume, max(0, drug2_pumped_volume))

    drugfree_medium_volume = culture.default_dilution_volume - drug1_pumped_volume  # - drug2_pumped_volume
    drugfree_medium_volume = min(culture.default_dilution_volume, max(0, drugfree_medium_volume))

    culture.dilute(pump1_volume=drugfree_medium_volume, pump2_volume=drug1_pumped_volume)



import time
import pyftdi.i2c


class Valves:
    open_duty_cycle = 0.03
    closed_duty_cycle = 0.12
    led_numbers = {valve: valve+7 for valve in [1, 2, 3, 4, 5, 6, 7, 8]}

    def __init__(self, device):
        self.device = device
        self.pwm_controller = device.pwm_controller
        self.is_open = {1: None,
                        2: None,
                        3: None,
                        4: None,
                        5: None,
                        6: None,
                        7: None}

        if self.device.is_connected():
            self.connect()

    def not_all_closed(self):
        return any(self.is_open.values())

    def connect(self):
        if self.pwm_controller.port is not None:
            self.idle_all()

    def set_duty_cycle_all(self, duty_cycle):
        self.pwm_controller.lock.acquire()
        try:
            self.pwm_controller.stop_all()
            for valve in range(1, 9):
                led_number = self.led_numbers[valve]
                self.pwm_controller.set_duty_cycle(led_number=led_number, duty_cycle=duty_cycle)
                time.sleep(0.5)
            self.pwm_controller.start_all()
        finally:
            self.pwm_controller.lock.release()

    def set_duty_cycle(self, valve, duty_cycle):
        """
        sets the duty cycle of the pwm signal for the valve.
        Stops pwm controller while changing value, to prevent the motor from
        moving after writing the first byte.
        """
        led_number = self.led_numbers[valve]
        self.pwm_controller.lock.acquire()
        try:
            self.pwm_controller.stop_all()
            self.pwm_controller.set_duty_cycle(led_number=led_number, duty_cycle=duty_cycle)
            self.pwm_controller.start_all()
        finally:
            self.pwm_controller.lock.release()

    def open(self, valve):
        self.set_duty_cycle(valve=valve, duty_cycle=self.open_duty_cycle)
        time.sleep(1.5)
        self.set_duty_cycle(valve=valve, duty_cycle=0)
        self.is_open[valve] = True

    def close(self, valve):
        self.set_duty_cycle(valve=valve, duty_cycle=self.closed_duty_cycle)
        time.sleep(1.5)
        self.set_duty_cycle(valve=valve, duty_cycle=0)
        self.is_open[valve] = False

    def open_all(self):
        for valve in range(1, 8):
            if not self.is_open[valve]:
                self.open(valve=valve)

        # self.set_duty_cycle_all(duty_cycle=self.open_duty_cycle)
        # time.sleep(1)  # wait for valves to move
        # self.set_duty_cycle_all(duty_cycle=0)

    def close_all(self):
        assert not self.device.is_pumping()
        for valve in range(1, 8):
            if self.is_open[valve] or self.is_open[valve] is None:
                self.close(valve=valve)

        # self.set_duty_cycle_all(duty_cycle=self.closed_duty_cycle)
        # time.sleep(1)  # wait for valves to move
        # self.set_duty_cycle_all(duty_cycle=0)

    def idle_all(self):
        self.set_duty_cycle_all(duty_cycle=0)

    def set_state(self, valve, is_open=False):
        assert is_open in [True, False]
        if is_open:
            self.open(valve)
        else:
            self.close(valve)



import pandas as pd
import numpy as np
import os
import time


def read_csv_tail(filepath, lines=10, _buffer=4096):
    """Tail a file and get X lines from the end
    modified from https://stackoverflow.com/questions/136168/get-last-n-lines-of-a-file-similar-to-tail/136368#136368
    """
    header = open(filepath).readline().rstrip().split(",")
    lines = int(lines)
    with open(filepath, "rb") as f:
        # place holder for the lines found
        lines_found = []

        # block counter will be multiplied by buffer
        # to get the block size from the end
        block_counter = -1

        # loop until we find X lines
        while len(lines_found) < lines+1:
            try:
                f.seek(block_counter * _buffer, os.SEEK_END)
            except IOError:  # either file is too small, or too many lines requested
                f.seek(0)
                lines_found = f.readlines()
                break
            lines_found = f.readlines()
            # we found enough lines, get out
            # decrement the block counter to get the
            # next X bytes
            block_counter -= 1

    lines_found = lines_found[1:]  # cut header if lines > total lines in file
    lines_found = lines_found[-lines:]
    lines_found = [l.decode().rstrip().replace(" ", "").split(",") for l in lines_found]
    lines_found = np.array(lines_found)
    data = lines_found[:, 1:]
    d=[]
    for data_line in data:
        try:
            d += [pd.to_numeric(data_line)]
        except ValueError:
            line = []
            for value in data_line:
                try:
                    line += [pd.to_numeric(value)]
                except ValueError:
                    line += [np.nan]
            d += line
    data = d

    data = [pd.to_numeric(data_line) for data_line in data]
    timestamp = lines_found[:, 0]
    df = pd.DataFrame(data, columns=header[1:])

    try:
        timestamp = pd.to_numeric(timestamp)    # 5 ms  for 10 000 lines
    except ValueError:
        timestamp = pd.to_datetime(timestamp)   # 1.3 seconds for 10 000 lines
        timestamp = pd.to_numeric(timestamp)

    df.index = timestamp

    df.index.name = header[0]
    return df


class CultureDict(dict):
    def __init__(self, device):
        super().__init__()
        self.device = device

    def __setitem__(self, vial, c):
        if c is not None:
            assert os.path.exists(self.device.directory)
            new_culture_directory = os.path.join(self.device.directory, "vial_%d" % vial)

            existing_config = os.path.join(new_culture_directory, "culture_config.yaml")
            if c.directory is None and os.path.exists(existing_config):
                raise RuntimeError("Culture %d config exists" % vial)
                # q = input("replace culture %d config?[y/n]" % vial)
                # assert q == "y"
            if not os.path.exists(new_culture_directory):
                os.mkdir(new_culture_directory)
            c.vial_number = vial
            # c.experiment_directory = self.device.directory
            c.directory = new_culture_directory
            c.device = self.device

        super().__setitem__(vial, c)


def write_variable(culture, variable_name, value):
    filepath = os.path.join(culture.directory, "%s.csv" % variable_name)
    if not os.path.exists(filepath):
        with open(filepath, "w+") as f:
            f.write("time,%s\n" % variable_name)
    with open(filepath, "a") as f:
        data_string = "%.1f, %.5f\n" % (int(time.time()), value)
        f.write(data_string)



import ipywidgets as widgets

from IPython.display import clear_output


class RunTab:
    def __init__(self, main_gui):
        self.main_gui = main_gui
        self.run_button = widgets.Button(description="RUN")
        self.stop_button = widgets.Button(description="STOP")
        if self.main_gui.experiment.is_running():
            self.run_button.disabled = True
        else:
            self.stop_button.disabled = True

        self.run_button.on_click(self.handle_run_button)
        self.widget = widgets.HBox([self.run_button, self.stop_button])

    def handle_run_button(self, b):
        with self.main_gui.status_bar.output:
            clear_output()
            self.main_gui.experiment.run()
        self.run_button.disabled = True
        self.stop_button.disabled = False

    def handle_stop_button(self, b):
        with self.main_gui.status_bar.output:
            clear_output()
            self.main_gui.experiment.stop()
        self.stop_button.disabled = True
        self.run_button.disabled = False







from src.replifactory import ProgramViewerTab
from src.replifactory import StatusBar
import ipywidgets as widgets
from ipywidgets import HBox
from src.replifactory import DeviceTab
from src.replifactory import DeviceConfigTab
from src.replifactory import CameraTab
from src.replifactory import ExperimentPlotTab
import os
from .experiment_status_tab import ExperimentStatusTab


class MainGui:
    def __init__(self):
        global _main_gui
        if "_main_gui" in globals().keys():

            q = input("initialize new GUI?[y/n]")
            if q == "y":
                _main_gui = self
            else:
                raise RuntimeError("Existing GUI was NOT overwritten by new GUI")
        self._device = None
        self._experiment = None

        self.status_bar = StatusBar(main_gui=self)

        self.device_tab = DeviceTab(main_gui=self)
        self.experiment_tab = ProgramViewerTab(main_gui=self)
        self.photo_tab = CameraTab()
        self.device_config_tab = DeviceConfigTab(main_gui=self)
        self.experiment_status_tab = ExperimentStatusTab(main_gui=self)
        self.experiment_plot_tab = ExperimentPlotTab(main_gui=self)
        self.tab_list = [self.experiment_tab, self.device_tab, self.device_config_tab,
                         self.experiment_status_tab, self.experiment_plot_tab, self.photo_tab]
        self.tabs = widgets.Tab()

        self.widget = HBox([self.status_bar.widget, self.tabs])
        self.update()
        _main_gui = self

    def update(self):
        self.status_bar.update()
        self.tabs.children = [tab.widget for tab in self.tab_list]
        for i in range(len(self.tab_list)):
            self.tabs.set_title(i, self.tab_list[i].title)

        self.widget = HBox([self.status_bar.widget, self.tabs])

    @property
    def device(self):
        if self.experiment is None:
            return self._device
        else:
            return self.experiment.device

    @device.setter
    def device(self, device):
        self._device = device
        if device is not None:
            self.device_config_tab.update()

    @property
    def experiment(self):
        return self._experiment

    @experiment.setter
    def experiment(self, experiment):
        self._experiment = experiment
        if self.experiment.device is not None:
            # self.device = self.experiment.device
            if self.device.__class__ in self.device_tab.device_class.options:
                self.device_tab.device_class.index = self.device_tab.device_class.options.index(self.device.__class__)
            else:
                print("Experiment config device class unknown.")
            if self.device.ftdi_address in self.device_tab.ftdi_address.options:
                self.device_tab.ftdi_address.index = self.device_tab.ftdi_address.options.index(self.device.ftdi_address)
            else:
                print("Experiment config FTDI address does not match connected devices.")
            config_paths = [p for p in self.device_tab.config_path.options if
                            os.path.realpath(p).startswith(os.path.realpath(self.device.directory))]
            if len(config_paths) > 0:
                self.device_tab.config_path.index = self.device_tab.config_path.options.index(config_paths[0])
            else:
                print("Experiment config device directory unknown.")
        else:
            self.experiment.device = self._device
            print("New device associated with experiment.")

# class Tabs:
#     def __init__(self, main_gui):
#         self.main_gui = main_gui
#         self.widget = widgets.Tab()
#         self.device_tab = DeviceTab(self.main_gui)
#         self.experiment_tab = ProgramViewerTab(experiment=self.main_gui.experiment)
#
#         self.tab_list = [self.device_tab, self.experiment_tab]
#         self.update()
#
#     def update(self):
#         self.widget.children = [tab.widget for tab in self.tab_list]
#         for i in range(len(self.tab_list)):
#             self.widget.set_title(i, self.tab_list[i].title)

    # self.widget.children = [DeviceControl(None).widget,
        #                         ProgramViewerTab(None).widget]
        # self.widget.children = [DeviceControl(device=self.main_gui.device).widget,
        #                         ProgramViewerTab(experiment=self.main_gui.experiment).widget]
        # self.widget.set_title(0, "Control")
        # self.widget.set_title(1, "Program viewer")


# class NewExperimentWidget:
#     def __init__(self):
#         textbox = widgets.Text(description="New experiment:", value="experiment name", continuous_update=False,
#                                style={'description_width': 'initial'})
#         self.widget = interactive(self.make_new_experiment, experiment_name=textbox)
#
#
#     def make_new_experiment(self, experiment_name):
#         if experiment_name != "experiment name":
#             if experiment_name not in os.listdir("."):
#                 os.mkdir(experiment_name)
#                 experiment_path = os.path.join(os.getcwd(), experiment_name)
#                 experiment_name = os.path.realpath(experiment_path)
#                 print("Initialized experiment in directory\n", experiment_name)
#             else:
#                 raise FileExistsError





import time
import os
import numpy as np
from scipy.optimize import curve_fit
from matplotlib import pyplot as plt


def measure_od_all(device, vials_to_measure=(1, 2, 3, 4, 5, 6, 7)):
    available_vials = []
    for v in vials_to_measure:
        if device.locks_vials[v].acquire(blocking=False):
            available_vials += [v]

    od_values = {}
    if len(available_vials) > 0:

        # Stop stirrers completely and wait 10 seconds for vortex to settle bubbles to raise above laser level.
        # Waiting enough time for the smallest bubbles to raise improves the measurement precision.
        # for vial in available_vials:
        #     device.stirrers.set_speed(vial=vial, speed=3)
        # time.sleep(2)
        # Stir at minimum speed (without forming vortex) to homogenize turbidity and prevent precipitation
        for vial in available_vials:
            device.stirrers.set_speed(vial=vial, speed=0)
        time.sleep(5)
        for vial in available_vials:
            device.stirrers.set_speed(vial=vial, speed=1)
        time.sleep(3)
        for vial in available_vials:
            device.stirrers.set_speed(vial=vial, speed=0)
        time.sleep(1)
        for vial in available_vials:
            od = device.od_sensors[vial].measure_od()
            od_values[vial] = od
        for vial in available_vials:
            device.stirrers.set_speed(vial=vial, speed=2)
        # time.sleep(2)
        # for vial in available_vials:
        #     device.stirrers.set_speed(vial=vial, speed=3)
        # time.sleep(5)
        # for vial in available_vials:
        #     device.stirrers.set_speed(vial=vial, speed=2)
        # for vial in available_vials:
        #     device.stirrers.set_speed(vial=vial, speed=3)
        # time.sleep(2)
        # for vial in available_vials:
        #     device.stirrers.set_speed(vial=vial, speed=2)
        # time.sleep(1)
        for v in available_vials:
            device.locks_vials[v].release()

        # assign values to culture.od parameter, which writes to csv file and calculates mu
        for v in od_values.keys():
            device.cultures[v].od = od_values[v]
    return od_values


def od_calibration_function(x, a, b, c, d, g):
    """
    4 Parameter logistic function
    adapted from:
    https://weightinginbayesianmodels.github.io/poctcalibration/calib_tut4_curve_background.html

    :param x: measured voltage in millivolts
    :param a: On the scale of y; horizontal asymptote as x goes to infinity.
    :param b: Hill coefficient
    :param c: Inflection point.
    :param d: On the scale of y; horizontal asymptote as x goes, in theory, to negative infinity, but, in practice, to zero.
    :return: y
    """
    y = a + (d-a)/((1+(x/c)**b)**g)
    return y


class OdSensor:
    def __init__(self, device, vial_number):
        self.device = device
        self.vial_number = vial_number
        self.calibration_function = None
        # self.fit_calibration_function()

    @property
    def calibration_mv_to_od(self):
        return self.device.calibration_mv_to_od[self.vial_number]

    @calibration_mv_to_od.setter
    def calibration_mv_to_od(self, value):
        self.device.calibration_mv_to_od[self.vial_number] = value
        self.fit_calibration_function()
        self.device.save()

    def fit_calibration_function(self):
        calibration_mv = list(self.device.calibration_mv_to_od[self.vial_number].keys())
        calibration_od = list(self.device.calibration_mv_to_od[self.vial_number].values())
        calibration_mv = np.array(sorted(calibration_mv))
        calibration_od = np.array(sorted(calibration_od)[::-1])
        drop_index = calibration_od < 5
        calibration_mv = calibration_mv[drop_index]
        calibration_od = calibration_od[drop_index]

        if calibration_mv is not None:
            if len(calibration_mv) >= 4:
                coefs, _ = curve_fit(od_calibration_function, calibration_mv, calibration_od, p0=(0, 2, 0, 20, 1),
                                     bounds=[(-0.5, 0, 0, 20, 0), (0.1, 5, 20, 200, 3)],
                                     sigma=1 / (calibration_mv * 0.1 + 0.0001), max_nfev=5000)
                a, b, c, d, g = coefs

                def f(mv):
                    """
                    converts mV to OD
                    """
                    return od_calibration_function(mv, a, b, c, d, g)
                self.calibration_function = f
                return a, b, c, d, g

    def add_calibration_point(self, mv, od):
        temp = self.calibration_mv_to_od
        temp[mv] = od
        self.calibration_mv_to_od = temp

    def calibration_curve(self):
        calibration_mv = list(self.calibration_mv_to_od.keys())
        calibration_od = list(self.calibration_mv_to_od.values())
        if len(calibration_mv) < 4:
            print("OD SENSOR %d: NOT ENOUGH CALIBRATION POINTS (%d/4)" % (self.vial_number, len(calibration_mv)))
        else:
            plt.figure(figsize=[6, 4], dpi=100)
            try:
                a, b, c, d, g = self.fit_calibration_function()
                xmin = min(calibration_mv)
                xmax = max(calibration_mv)
                x = np.linspace(xmin, 50, 501)
                y = self.calibration_function(x)
                plt.plot(x, y, "k:")
                plt.title("od_min: %.3f    slope: %.3f    mv_inflec: %.3f    od_max: %.3f, g: %.3f" % (a, b, c, d, g), fontsize=8)
            except:
                print("Could not fit calibration function")

            plt.plot(calibration_mv, calibration_od, "r.")
            plt.xlabel("ADC signal (mV)")
            plt.ylabel("Optical Density")
            plt.suptitle("OD sensor calibration, vial %d" % self.vial_number, fontsize=10)
            # plt.show()

    def measure_transmitted_intensity(self):
        self.device.photodiodes.switch_to_vial(vial=self.vial_number)
        self.device.lasers.switch_on(vial=self.vial_number)
        time.sleep(0.02)
        mv, err = self.device.photodiodes.measure(gain=8, bitrate=16)
        self.device.lasers.switch_off(vial=self.vial_number)
        return mv, err

    def measure_background_intensity(self):
        self.device.photodiodes.switch_to_vial(vial=self.vial_number)
        mv, err = self.device.photodiodes.measure(gain=8, bitrate=16)
        return mv, err

    # def pause_mixers_and_measure_od(self):
    #     vial = self.vial_number
    #     self.device.locks_vials[vial].acquire()
    #     while self.device.is_pumping_in_vial == vial:
    #         time.sleep(1)
    #
    #     self.device.stirrers.set_speed(vial=vial, speed=0)  # wait 10s for vortex to stop
    #     time.sleep(10)
    #     self.device.stirrers.set_speed(vial=vial, speed=1)  # mix very slowly while measuring OD
    #     time.sleep(5)
    #
    #     od = self.measure_od()
    #
    #     self.device.stirrers.set_speed(vial=vial, speed=2)  # resume normal stirring speed
    #     self.device.locks_vials[vial].release()
    #     return od

    # def log_od(self, od):
    #     directory = os.path.join(self.device.directory, "vial_%d" % self.vial_number)
    #     if not os.path.exists(directory):
    #         os.mkdir(directory)
    #
    #     filepath = os.path.join(directory, "od.csv")
    #     if not os.path.exists(filepath):
    #         with open(filepath, "w+") as f:
    #             f.write("time,od\n")
    #     with open(filepath, "a") as f:
    #         data_string = "%.1f,%.4f\n" % (time.time(), od)
    #         f.write(data_string)

    def log_mv(self, background, transmitted):
        directory = os.path.join(self.device.directory, "vial_%d" % self.vial_number)
        if not os.path.exists(directory):
            os.mkdir(directory)

        filepath = os.path.join(directory, "photodiode_millivolts.csv")
        if not os.path.exists(filepath):
            with open(filepath, "w+") as f:
                f.write("time,transmitted,background\n")
        with open(filepath, "a") as f:
            data_string = "%d,%.4f,%.4f\n" % (int(time.time()), transmitted, background)
            f.write(data_string)

    def measure_signal(self):
        background = self.measure_background_intensity()[0]
        transmitted = self.measure_transmitted_intensity()[0]
        if self.device.directory is not None:
            self.log_mv(background=background, transmitted=transmitted)
        signal = transmitted - background
        return signal

    def measure_od(self):
        signal = self.measure_signal()
        if not callable(self.calibration_function):
            self.fit_calibration_function()

        od = self.calibration_function(signal)
        # self.log_od(od=od)  # logging in the culture class also updates growth rate
        return od

    def check(self):
        assert self.calibration_function is not None



import ipywidgets as widgets
from ipywidgets import VBox, HBox
import time
import os


class CameraTab:
    def __init__(self):
        self.title = "Camera"
        self.button = widgets.Button(description="photo", icon="camera")
        output = widgets.Output()
        img_widget = widgets.Image(format='png', width=600, height=400)
        self.vflip = widgets.Checkbox(value=False, description="vertical flip")
        self.hflip = widgets.Checkbox(value=False, description="horizontal flip")

        def on_button_clicked(b):
            self.button.disabled = True
            self.button.description = "busy"
            additional_args = ""
            if self.vflip.value:
                additional_args += " --vflip"
            if self.hflip.value:
                additional_args += " --hflip"

            res = os.system("raspistill -o photo.png %s" % additional_args)
            if res == 0:
                img_widget.value = open("./photo.png", "rb").read()
                with output:
                    output.clear_output()
                    print(time.ctime())
            else:
                with output:
                    output.clear_output()
                    print("raspistill command not working.")

            self.button.disabled = False
            self.button.description = "photo"

        self.button.on_click(on_button_clicked)
        self.widget = VBox([HBox([self.button, self.vflip, self.hflip]), output, img_widget])
